"""rio_tiler.mosaic.backend: base Backend class."""

import abc
import logging
import warnings
from typing import Any

import attr
import numpy
from morecantile import TileMatrixSet
from pydantic import BaseModel, ConfigDict, Field
from rasterio.crs import CRS
from rasterio.errors import NotGeoreferencedWarning
from rasterio.features import bounds as featureBounds
from rasterio.features import rasterize
from rasterio.warp import transform_geom

from rio_tiler.constants import WEB_MERCATOR_TMS, WGS84_CRS
from rio_tiler.errors import NoAssetFoundError, PointOutsideBounds
from rio_tiler.io import BaseReader, MultiBaseReader, Reader
from rio_tiler.models import BandStatistics, ImageData, PointData
from rio_tiler.mosaic import mosaic_reader
from rio_tiler.tasks import multi_values_list
from rio_tiler.types import BBox
from rio_tiler.utils import (
    CRS_to_uri,
    Timer,
    _validate_shape_input,
    inherit_rasterio_env,
)

logger = logging.getLogger(__name__)


class MosaicInfo(BaseModel):
    """Mosaic info responses."""

    bounds: BBox = Field(default=(-180, -90, 180, 90))
    crs: str

    model_config = ConfigDict(extra="allow")


@attr.s
class BaseBackend(BaseReader):
    """Base Class for mosaic backend.

    Attributes:
        input (str): mosaic path.
        tms (morecantile.TileMatrixSet, optional): TileMatrixSet grid definition. Defaults to `WebMercatorQuad`.
        reader (rio_tiler.io.BaseReader): Dataset reader. Defaults to `rio_tiler.io.Reader`.
        reader_options (dict): Options to forward to the reader config.
        bounds (tuple): mosaic bounds (left, bottom, right, top). **READ ONLY attribute**.
        crs (rasterio.crs.CRS): mosaic crs in which its bounds is defined. **READ ONLY attribute**.
        minzoom (int): mosaic minimum zoom level. **READ ONLY attribute**.
        maxzoom (int): mosaic maximum zoom level. **READ ONLY attribute**.

    """

    input: str = attr.ib()
    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    reader: type[BaseReader] | type[MultiBaseReader] = attr.ib(default=Reader)
    reader_options: dict = attr.ib(factory=dict)

    bounds: BBox = attr.ib(init=False)
    crs: CRS = attr.ib(init=False)

    minzoom: int = attr.ib(init=False)
    maxzoom: int = attr.ib(init=False)

    @abc.abstractmethod
    def assets_for_tile(self, x: int, y: int, z: int, **kwargs: Any) -> list[Any]:
        """Retrieve assets for tile."""

    @abc.abstractmethod
    def assets_for_point(
        self,
        lng: float,
        lat: float,
        coord_crs: CRS | None = None,
        **kwargs: Any,
    ) -> list[Any]:
        """Retrieve assets for point."""

    @abc.abstractmethod
    def assets_for_bbox(
        self,
        xmin: float,
        ymin: float,
        xmax: float,
        ymax: float,
        coord_crs: CRS | None = None,
        **kwargs,
    ) -> list[Any]:
        """Retrieve assets for bbox."""

    def asset_name(self, asset: Any) -> str:
        """Get asset name."""
        return str(asset)

    def info(self) -> MosaicInfo:  # type: ignore
        """Mosaic info."""
        return MosaicInfo(
            bounds=self.bounds,
            crs=CRS_to_uri(self.crs) or self.crs.to_wkt(),
        )

    def point(  # type: ignore
        self,
        lon: float,
        lat: float,
        coord_crs: CRS = WGS84_CRS,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> list[tuple[str, PointData]]:
        """Get Point value from multiple assets."""
        search_options = search_options or {}
        mosaic_assets = self.assets_for_point(
            lon, lat, coord_crs=coord_crs, **search_options
        )
        if not mosaic_assets:
            raise NoAssetFoundError(f"No assets found for point ({lon},{lat})")

        @inherit_rasterio_env
        def _reader(
            asset: Any, lon: float, lat: float, coord_crs: CRS, **kwargs
        ) -> PointData:
            with self.reader(asset, **self.reader_options) as src_dst:
                return src_dst.point(lon, lat, coord_crs=coord_crs, **kwargs)

        if "allowed_exceptions" not in kwargs:
            kwargs.update({"allowed_exceptions": (PointOutsideBounds,)})

        logger.info(
            f"reading Point for {len(mosaic_assets)} assets with reader: {self.reader}"
        )
        return [
            (self.asset_name(asset), pt)
            for asset, pt in multi_values_list(
                mosaic_assets, _reader, lon, lat, coord_crs, **kwargs
            )
        ]

    def tile(  # type: ignore
        self,
        x: int,
        y: int,
        z: int,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> tuple[ImageData, list[str]]:
        """Get Tile from multiple assets."""
        timings = []
        with Timer() as t:
            search_options = search_options or {}
            mosaic_assets = self.assets_for_tile(x, y, z, **search_options)
        timings.append(("search", round(t.elapsed * 1000, 2)))

        if not mosaic_assets:
            raise NoAssetFoundError(f"No assets found for tile {z}-{x}-{y}")

        @inherit_rasterio_env
        def _reader(asset: Any, x: int, y: int, z: int, **kwargs: Any) -> ImageData:
            with self.reader(asset, tms=self.tms, **self.reader_options) as src_dst:
                return src_dst.tile(x, y, z, **kwargs)

        logger.info(
            f"reading Tile for {len(mosaic_assets)} assets with reader: {self.reader}"
        )
        with Timer() as t:
            img, asset_used = mosaic_reader(mosaic_assets, _reader, x, y, z, **kwargs)
        timings.append(("mosaicking", round(t.elapsed * 1000, 2)))
        img.metadata = {**img.metadata, "timings": timings}

        asset_used = [self.asset_name(asset) for asset in asset_used]
        return img, asset_used

    def part(  # type: ignore
        self,
        bbox: BBox,
        bounds_crs: CRS = WGS84_CRS,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> tuple[ImageData, list[str]]:
        """Create an Image from multiple assets for a bbox."""
        xmin, ymin, xmax, ymax = bbox
        timings = []

        with Timer() as t:
            search_options = search_options or {}
            mosaic_assets = self.assets_for_bbox(
                xmin,
                ymin,
                xmax,
                ymax,
                coord_crs=bounds_crs,
                **search_options,
            )
        timings.append(("search", round(t.elapsed * 1000, 2)))

        if not mosaic_assets:
            raise NoAssetFoundError("No assets found for bbox input")

        @inherit_rasterio_env
        def _reader(asset: Any, bbox: BBox, bounds_crs: CRS, **kwargs: Any) -> ImageData:
            with self.reader(asset, **self.reader_options) as src_dst:
                return src_dst.part(bbox, bounds_crs=bounds_crs, **kwargs)

        logger.info(
            f"reading Part for {len(mosaic_assets)} assets with reader: {self.reader}"
        )
        with Timer() as t:
            img, asset_used = mosaic_reader(
                mosaic_assets, _reader, bbox, bounds_crs, **kwargs
            )
        timings.append(("mosaicking", round(t.elapsed * 1000, 2)))
        img.metadata = {**img.metadata, "timings": timings}

        asset_used = [self.asset_name(asset) for asset in asset_used]
        return img, asset_used

    def feature(  # type: ignore
        self,
        shape: dict,
        shape_crs: CRS = WGS84_CRS,
        dst_crs: CRS | None = None,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> tuple[ImageData, list[str]]:
        """Create an Image from multiple assets for a GeoJSON feature."""
        shape = _validate_shape_input(shape)

        if not dst_crs:
            dst_crs = shape_crs

        bbox = featureBounds(shape)

        img, asset_used = self.part(
            bbox,
            bounds_crs=shape_crs,
            dst_crs=dst_crs,
            search_options=search_options,
            **kwargs,
        )

        if dst_crs != shape_crs:
            shape = transform_geom(shape_crs, dst_crs, shape)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=NotGeoreferencedWarning,
                module="rasterio",
            )
            cutline_mask = rasterize(
                [shape],
                out_shape=(img.height, img.width),
                transform=img.transform,
                all_touched=True,  # Mandatory for matching masks at different resolutions
                default_value=0,
                fill=1,
                dtype="uint8",
            ).astype("bool")

        img.array.mask = numpy.where(~cutline_mask, img.array.mask, True)

        return img, asset_used

    ############################################################################
    # Not Implemented methods
    # BaseReader required those method to be implemented
    def statistics(
        self,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> dict[str, BandStatistics]:
        """PlaceHolder for statistics."""
        raise NotImplementedError

    def preview(  # type: ignore
        self,
        search_options: dict | None = None,
        **kwargs: Any,
    ) -> tuple[ImageData, list[str]]:
        """PlaceHolder for preview."""
        raise NotImplementedError



"""rio-tiler models."""

import itertools
import re
import warnings
from collections.abc import Sequence
from typing import Any, Literal, Self, cast

import attr
import numpy
import rasterio
from affine import Affine
from color_operations import parse_operations, scale_dtype, to_math_type
from numpy.typing import NDArray
from pydantic import BaseModel
from rasterio import windows
from rasterio.coords import BoundingBox
from rasterio.crs import CRS
from rasterio.dtypes import dtype_ranges
from rasterio.enums import ColorInterp, Resampling
from rasterio.errors import NotGeoreferencedWarning
from rasterio.features import rasterize
from rasterio.io import MemoryFile
from rasterio.plot import reshape_as_image
from rasterio.transform import array_bounds, from_bounds
from rasterio.warp import calculate_default_transform, reproject, transform_geom

from rio_tiler.colormap import apply_cmap
from rio_tiler.constants import WGS84_CRS
from rio_tiler.errors import (
    InvalidDatatypeWarning,
    InvalidExpression,
    InvalidPointDataError,
)
from rio_tiler.expression import apply_expression, get_expression_blocks
from rio_tiler.types import (
    BBox,
    ColorMapType,
    GDALColorMapType,
    IntervalTuple,
    NoData,
    RIOResampling,
    WarpResampling,
)
from rio_tiler.utils import (
    _validate_shape_input,
    get_array_statistics,
    linear_rescale,
    non_alpha_indexes,
    render,
    resize_array,
)


class Bounds(BaseModel):
    """Dataset Bounding box"""

    bounds: BBox
    crs: str


class Info(Bounds):
    """Dataset Info."""

    band_metadata: list[tuple[str, dict]]
    band_descriptions: list[tuple[str, str]]
    dtype: str
    nodata_type: Literal["Alpha", "Mask", "Internal", "Nodata", "None"]
    colorinterp: list[str] | None = None
    scales: list[float] | None = None
    offsets: list[float] | None = None
    colormap: GDALColorMapType | None = None

    model_config = {"extra": "allow"}


class BandStatistics(BaseModel):
    """Band statistics"""

    min: float
    max: float
    mean: float
    count: float
    sum: float
    std: float
    median: float
    majority: float
    minority: float
    unique: float
    histogram: list[list[float | int]]
    valid_percent: float
    masked_pixels: float
    valid_pixels: float
    description: str

    model_config = {"extra": "allow"}


def to_coordsbbox(bbox) -> BoundingBox | None:
    """Convert bbox to CoordsBbox nameTuple."""
    return BoundingBox(*bbox) if bbox else None


def rescale_image(
    array: numpy.ma.MaskedArray,
    in_range: Sequence[IntervalTuple],
    out_range: Sequence[IntervalTuple] = ((0, 255),),
    out_dtype: str | numpy.number = "uint8",
) -> numpy.ma.MaskedArray:
    """Rescale image data in-place."""
    if len(array.shape) < 3:
        array = numpy.expand_dims(array, axis=0)  # type: ignore [assignment]

    nbands = array.shape[0]

    if len(in_range) != nbands:
        in_range = ((in_range[0]),) * nbands

    if len(out_range) != nbands:
        out_range = ((out_range[0]),) * nbands

    for bdx in range(nbands):
        array.data[bdx] = numpy.where(
            ~array.mask[bdx],  # type: ignore [index]
            linear_rescale(
                array.data[bdx], in_range=in_range[bdx], out_range=out_range[bdx]
            ),
            0,
        )

    return array.astype(out_dtype)  # type: ignore [return-value]


def to_masked(array: numpy.ndarray) -> numpy.ma.MaskedArray:
    """Makes sure we have a MaskedArray."""
    if not numpy.ma.isarray(array):
        array = numpy.ma.asarray(array)

    array = cast(numpy.ma.MaskedArray, array)
    # when a masked array is totally valid, the mask is set to numpy.ma.nomask
    # https://numpy.org/doc/stable/reference/maskedarray.baseclass.html#numpy.ma.nomask
    # doing `array.mask = False` force the creation of the mask array
    if not array.mask.shape:
        array.mask = False

    return array


@attr.s
class PointData:
    """Point Data class.

    Attributes:
        array (numpy.ma.MaskedArray): pixel values.
        band_names (list): name of each band. Defaults to `["1", "2", "3"]` for 3 bands image.
        coordinates (tuple): Point's coordinates.
        crs (rasterio.crs.CRS, optional): Coordinates Reference System of the bounds.
        assets (list, optional): list of assets used to construct the data values.
        metadata (dict, optional): Additional metadata. Defaults to `{}`.

    """

    array: numpy.ma.MaskedArray = attr.ib(converter=to_masked)
    band_names: list[str] = attr.ib(kw_only=True)
    band_descriptions: list[str] = attr.ib(kw_only=True)
    coordinates: tuple[float, float] | None = attr.ib(default=None, kw_only=True)
    crs: CRS | None = attr.ib(default=None, kw_only=True)
    assets: list | None = attr.ib(default=None, kw_only=True)
    metadata: dict | None = attr.ib(factory=dict, kw_only=True)
    nodata: NoData | None = attr.ib(default=None, kw_only=True)
    scales: list[float | int] = attr.ib(kw_only=True)
    offsets: list[float | int] = attr.ib(kw_only=True)
    pixel_location: tuple[float | int, float | int] | None = attr.ib(
        default=None, kw_only=True
    )

    @array.validator
    def _validate_data(self, attribute, value):
        """PointsData data has to be a 1d array."""
        if not len(value.shape) == 1:
            raise ValueError("PointsData data has to be a 1D array")

    @coordinates.validator
    def _validate_coordinates(self, attribute, value):
        """coordinates has to be a 2d list."""
        if value and not len(value) == 2:
            raise ValueError("Coordinates data has to be a 2d list")

    @band_names.default
    def _default_band_names(self):
        return [f"b{ix + 1}" for ix in range(self.count)]

    @band_descriptions.default
    def _default_band_descriptions(self):
        return ["" or f"b{ix + 1}" for ix in range(self.count)]

    @scales.default
    def _default_scales(self):
        return [1.0] * self.count

    @offsets.default
    def _default_offsets(self):
        return [0.0] * self.count

    @property
    def data(self) -> numpy.ndarray:
        """Return data part of the masked array."""
        return self.array.data

    @property
    def _mask(self):
        """Return `inverted/merged` mask from data array."""
        return numpy.array([numpy.logical_and.reduce(~self.array.mask)])

    @property
    def mask(self) -> numpy.ndarray:
        """Return Mask in form of rasterio dataset mask."""
        minv, maxv = dtype_ranges[str(self.array.dtype)]
        return numpy.where(self._mask, maxv, minv).astype(self.array.dtype)

    def __iter__(self):
        """Allow for variable expansion."""
        for i in self.array.data:
            yield i

    @property
    def count(self) -> int:
        """Number of band."""
        return self.array.shape[0]

    @classmethod
    def create_from_list(cls, data: Sequence["PointData"]) -> "PointData":
        """Create PointData from a sequence of PointsData objects.

        Args:
            data (sequence): sequence of PointData.

        """
        if not data:
            raise InvalidPointDataError("Empty PointData list.")

        # validate coordinates
        if all(pt.coordinates or pt.crs or None for pt in data):
            lon, lat, crs = zip(*[(*(pt.coordinates or []), pt.crs) for pt in data])
            if len(set(lon)) > 1 or len(set(lat)) > 1 or len(set(crs)) > 1:
                raise InvalidPointDataError(
                    "Cannot concatenate points with different coordinates/CRS."
                )

        arr = numpy.ma.concatenate([pt.array for pt in data])

        assets = list(
            dict.fromkeys(
                itertools.chain.from_iterable([pt.assets for pt in data if pt.assets])
            )
        )

        band_names = list(
            itertools.chain.from_iterable([pt.band_names for pt in data if pt.band_names])
        )

        band_descriptions = list(
            itertools.chain.from_iterable(
                [pt.band_descriptions for pt in data if pt.band_descriptions]
            )
        )

        scales = list(itertools.chain.from_iterable([pt.scales for pt in data]))

        offsets = list(itertools.chain.from_iterable([pt.offsets for pt in data]))

        metadata = dict(
            itertools.chain.from_iterable(
                [pt.metadata.items() for pt in data if pt.metadata]
            )
        )

        return cls(
            arr,
            assets=assets,
            band_names=band_names,
            band_descriptions=band_descriptions,
            offsets=offsets,
            scales=scales,
            coordinates=data[0].coordinates,
            crs=data[0].crs,
            metadata=metadata,
        )

    def apply_expression(self, expression: str) -> "PointData":
        """Apply expression to the image data."""
        if "eval(" in expression:
            raise InvalidExpression("Invalid expression.")

        blocks = get_expression_blocks(expression)

        data = apply_expression(blocks, self.band_names, self.array)
        # Using numexpr do not preserve mask info
        data.mask = False

        mapexpr = {
            self.band_names[idx]: desc for idx, desc in enumerate(self.band_descriptions)
        }
        _re = re.compile(r"\bb[0-9A-Z]+\b")
        band_descriptions = [
            _re.sub(lambda x: mapexpr[x.group()], block) for block in blocks
        ]

        return PointData(
            data,
            assets=self.assets,
            crs=self.crs,
            coordinates=self.coordinates,
            band_descriptions=band_descriptions,
            metadata=self.metadata,
            pixel_location=self.pixel_location,
        )


def masked_and_3d(array: numpy.ndarray) -> numpy.ma.MaskedArray:
    """Makes sure we have a 3D array and mask"""
    if not numpy.ma.isarray(array):
        array = numpy.ma.asarray(array)

    if len(array.shape) < 3:
        array = numpy.expand_dims(array, axis=0)

    array = cast(numpy.ma.MaskedArray, array)

    # when a masked array is totally valid, the mask is set to numpy.ma.nomask
    # https://numpy.org/doc/stable/reference/maskedarray.baseclass.html#numpy.ma.nomask
    # doing `array.mask = False` force the creation of the mask array
    if not array.mask.shape:
        array.mask = False

    return array


@attr.s
class ImageData:
    """Image Data class.

    Attributes:
        array (numpy.ma.MaskedArray): image values.
        assets (list, optional): list of assets used to construct the data values.
        bounds (BoundingBox, optional): bounding box of the data.
        crs (rasterio.crs.CRS, optional): Coordinates Reference System of the bounds.
        metadata (dict, optional): Additional metadata. Defaults to `{}`.
        band_names (list, optional): name of each band. Defaults to `["b1", "b2", "b3"]` for 3 bands image.
        dataset_statistics (list, optional): dataset statistics `[(min, max), (min, max)]`

    Note: `mask` should be considered as `PER_BAND` so shape should be similar as the data

    """

    array: numpy.ma.MaskedArray = attr.ib(converter=masked_and_3d)
    assets: list | None = attr.ib(default=None, kw_only=True)
    bounds: BoundingBox | None = attr.ib(
        default=None, converter=to_coordsbbox, kw_only=True
    )
    crs: CRS | None = attr.ib(default=None, kw_only=True)
    metadata: dict | None = attr.ib(factory=dict, kw_only=True)
    nodata: NoData | None = attr.ib(default=None, kw_only=True)
    scales: list[float | int] = attr.ib(kw_only=True)
    offsets: list[float | int] = attr.ib(kw_only=True)
    band_names: list[str] = attr.ib(kw_only=True)
    band_descriptions: list[str] = attr.ib(kw_only=True)
    dataset_statistics: Sequence[tuple[float, float]] | None = attr.ib(
        default=None, kw_only=True
    )
    cutline_mask: numpy.ndarray | None = attr.ib(default=None)
    alpha_mask: numpy.ndarray | None = attr.ib(default=None)

    @band_names.default
    def _default_band_names(self):
        return [f"b{ix + 1}" for ix in range(self.count)]

    @band_descriptions.default
    def _default_band_descriptions(self):
        return ["" or f"b{ix + 1}" for ix in range(self.count)]

    @scales.default
    def _default_scales(self):
        return [1.0] * self.count

    @offsets.default
    def _default_offsets(self):
        return [0.0] * self.count

    @alpha_mask.validator
    def _check_alpha_mask(self, attribute, value):
        """Make sure alpha mask has valid shame and datatype."""
        if value is not None:
            if (
                len(value.shape) != 2
                or value.shape[0] != self.height
                or value.shape[1] != self.width
            ):
                raise ValueError(
                    f"Invalide shape {value.shape} for AlphaMask, should be of shape {self.height}x{self.width}"
                )

            if not value.dtype == self.array.dtype:
                raise ValueError(
                    f"Invalide dtype {value.dtype} for AlphaMask, should be of {self.array.dtype}"
                )

    @property
    def data(self) -> numpy.ndarray:
        """Return data part of the masked array."""
        return self.array.data

    @property
    def mask(self) -> numpy.ndarray:
        """Return Mask in form of rasterio dataset mask."""
        # NOTE: if available we return the alpha_mask
        if self.alpha_mask is not None:
            return self.alpha_mask

        minv, maxv = dtype_ranges[str(self.array.dtype)]
        return numpy.where(self._mask, maxv, minv).astype(self.array.dtype)

    @property
    def _mask(self):
        """Return `inverted/merged` mask from data array."""
        return numpy.logical_or.reduce(~self.array.mask)

    def __iter__(self):
        """Allow for variable expansion (``arr, mask = ImageData``)"""
        for i in (self.array.data, self.mask):
            yield i

    @classmethod
    def from_bytes(cls, data: bytes) -> "ImageData":
        """Create ImageData from bytes.

        Args:
            data (bytes): raster dataset as bytes.

        """
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=NotGeoreferencedWarning,
                module="rasterio",
            )
            with MemoryFile(data) as m:
                with m.open() as dataset:
                    indexes = non_alpha_indexes(dataset)
                    if ColorInterp.alpha in dataset.colorinterp:
                        alpha_idx = dataset.colorinterp.index(ColorInterp.alpha) + 1
                        idx = tuple(indexes) + (alpha_idx,)
                        array = dataset.read(indexes=idx)

                        mask = ~array[-1].astype("bool")
                        array = numpy.ma.MaskedArray(array[0:-1])
                        array.mask = mask

                    else:
                        array = dataset.read(indexes=indexes, masked=True)

                    stats = []
                    for ix in indexes:
                        tags = dataset.tags(ix)
                        if all(
                            stat in tags
                            for stat in ["STATISTICS_MINIMUM", "STATISTICS_MAXIMUM"]
                        ):
                            stat_min = float(tags.get("STATISTICS_MINIMUM"))
                            stat_max = float(tags.get("STATISTICS_MAXIMUM"))
                            stats.append((stat_min, stat_max))

                    dataset_statistics = stats if len(stats) == len(indexes) else None

                    return cls(
                        array,
                        crs=dataset.crs,
                        bounds=dataset.bounds,
                        band_names=[f"b{idx}" for idx in indexes],
                        band_descriptions=[
                            dataset.descriptions[ix - 1] or "" for idx in indexes
                        ],
                        dataset_statistics=dataset_statistics,
                        nodata=dataset.nodata,
                        scales=list(dataset.scales),
                        offsets=list(dataset.offsets),
                        metadata=dataset.tags(),
                    )

    @classmethod
    def create_from_list(cls, data: Sequence["ImageData"]) -> "ImageData":
        """Create ImageData from a sequence of ImageData objects.

        Args:
            data (sequence): sequence of ImageData.

        """
        h, w = zip(*[(img.height, img.width) for img in data])

        # Get cutline mask at highest resolution.
        max_h, max_w = max(h), max(w)
        cutline_mask = next(
            img.cutline_mask for img in data if img.height == max_h and img.width == max_w
        )

        if len(set(h)) > 1 or len(set(w)) > 1:
            warnings.warn(
                "Cannot concatenate images with different sizes. Will resize using max width/height.",
                UserWarning,
            )
            for img in data:
                if img.height == max_h and img.width == max_w:
                    continue
                arr = numpy.ma.MaskedArray(
                    resize_array(img.array.data, max_h, max_w),  # type: ignore [arg-type]
                    mask=resize_array(img.array.mask * 1, max_h, max_w).astype("bool"),  # type: ignore [arg-type]
                )
                img.array = arr

        arr = numpy.ma.concatenate([img.array for img in data])

        assets = list(
            dict.fromkeys(
                itertools.chain.from_iterable([img.assets for img in data if img.assets])
            )
        )

        bounds_values = [img.bounds for img in data if img.bounds]
        bounds = bounds_values[0] if bounds_values else None

        crs_values = [img.crs for img in data if img.crs]
        crs = crs_values[0] if crs_values else None

        band_names = list(
            itertools.chain.from_iterable(
                [img.band_names for img in data if img.band_names]
            )
        )

        band_descriptions = list(
            itertools.chain.from_iterable(
                [img.band_descriptions for img in data if img.band_descriptions]
            )
        )

        scales = list(itertools.chain.from_iterable([img.scales for img in data]))

        offsets = list(itertools.chain.from_iterable([img.offsets for img in data]))

        stats = list(
            itertools.chain.from_iterable(
                [img.dataset_statistics for img in data if img.dataset_statistics]
            )
        )
        dataset_statistics = stats if len(stats) == len(band_names) else None

        metadata = dict(
            itertools.chain.from_iterable(
                [img.metadata.items() for img in data if img.metadata]
            )
        )

        return cls(
            arr,
            assets=assets,
            crs=crs,
            bounds=bounds,
            band_names=band_names,
            band_descriptions=band_descriptions,
            dataset_statistics=dataset_statistics,
            cutline_mask=cutline_mask,
            metadata=metadata,
            scales=scales,
            offsets=offsets,
        )

    def data_as_image(self) -> numpy.ndarray:
        """Return the data array reshaped into an image processing/visualization software friendly order.

        (bands, rows, columns) -> (rows, columns, bands).

        """
        return reshape_as_image(self.array)

    @property
    def width(self) -> int:
        """Width of the data array."""
        return self.array.shape[2]

    @property
    def height(self) -> int:
        """Height of the data array."""
        return self.array.shape[1]

    @property
    def count(self) -> int:
        """Number of band."""
        return self.array.shape[0]

    @property
    def transform(self) -> Affine:
        """Returns the affine transform."""
        return (
            from_bounds(*self.bounds, self.width, self.height)
            if self.bounds
            else Affine.scale(self.width, -self.height)
        )

    def rescale(
        self,
        in_range: Sequence[IntervalTuple],
        out_range: Sequence[IntervalTuple] = ((0, 255),),
        out_dtype: str | numpy.number = "uint8",
    ) -> Self:
        """Rescale data in place."""
        self.array = rescale_image(
            self.array,
            in_range=in_range,
            out_range=out_range,
            out_dtype=out_dtype,
        )
        if self.alpha_mask is not None:
            self.alpha_mask = linear_rescale(
                self.alpha_mask,
                in_range=dtype_ranges[str(self.alpha_mask.dtype)],
                out_range=dtype_ranges[out_dtype],
            ).astype(out_dtype)

        # reset scales/offsets
        self.scales = [1.0] * self.count
        self.offsets = [0.0] * self.count

        return self

    def apply_color_formula(self, color_formula: str | None) -> Self:
        """Apply color-operations formula in place."""
        out = self.array.data
        out[out < 0] = 0

        for ops in parse_operations(color_formula):
            out = scale_dtype(ops(to_math_type(out)), numpy.uint8)

        data = numpy.ma.MaskedArray(out)
        data.mask = self.array.mask
        self.array = data

        # NOTE: we need to rescale the alpha mask if not in Uint8
        if self.alpha_mask is not None and self.alpha_mask.dtype != "uint8":
            self.alpha_mask = linear_rescale(
                self.alpha_mask, in_range=dtype_ranges[str(self.alpha_mask.dtype)]
            ).astype("uint8")

        # reset scales/offsets
        self.scales = [1.0] * self.count
        self.offsets = [0.0] * self.count

        return self

    def apply_colormap(self, colormap: ColorMapType) -> "ImageData":
        """Apply colormap to the image data."""
        data, alpha = apply_cmap(self.array.data, colormap)

        array = numpy.ma.MaskedArray(data)

        # `inValid/Valid` values for colormaped datatype (e.g 0 for uint8)
        invalid, valid = dtype_ranges[str(alpha.dtype)]

        # Combine both dataset mask and alpha from colormap
        array.mask = numpy.bitwise_or(alpha != valid, ~self._mask)

        if self.alpha_mask is not None:
            warnings.warn(
                "Alpha Mask value ignored from the input ImageData object when applying colormap.",
                UserWarning,
            )

        return ImageData(
            array,
            assets=self.assets,
            crs=self.crs,
            bounds=self.bounds,
            metadata=self.metadata,
            # NOTE: make sure masked part from initial data are also masked in alpha band
            alpha_mask=numpy.where(~self._mask, invalid, alpha),
        )

    def apply_expression(self, expression: str) -> "ImageData":
        """Apply expression to the image data."""
        if "eval(" in expression:
            raise InvalidExpression("Invalid expression.")

        blocks = get_expression_blocks(expression)

        stats = self.dataset_statistics
        if stats:
            res = []
            for prod in itertools.product(*stats):  # type: ignore
                res.append(apply_expression(blocks, self.band_names, numpy.array(prod)))

            stats = list(
                zip(
                    [float(min(r)) for r in zip(*res)],
                    [float(max(r)) for r in zip(*res)],
                )
            )

        data = apply_expression(blocks, self.band_names, self.array.copy())
        # NOTE: We use dataset mask when mixing bands
        data.mask = numpy.logical_or.reduce(self.array.mask)

        mapexpr = {
            self.band_names[idx]: desc for idx, desc in enumerate(self.band_descriptions)
        }
        _re = re.compile(r"\bb[0-9A-Z]+\b")
        band_descriptions = [
            _re.sub(lambda x: mapexpr[x.group()], block) for block in blocks
        ]

        return ImageData(
            data,
            assets=self.assets,
            crs=self.crs,
            bounds=self.bounds,
            band_descriptions=band_descriptions,
            metadata=self.metadata,
            dataset_statistics=stats,
            alpha_mask=self.alpha_mask.astype(data.dtype)
            if self.alpha_mask is not None
            else None,
        )

    def resize(
        self,
        height: int,
        width: int,
        resampling_method: RIOResampling = "nearest",
    ) -> "ImageData":
        """Resize data and mask."""
        data = resize_array(self.array.data, height, width, resampling_method)  # type: ignore [arg-type]
        mask = resize_array(self.array.mask * 1, height, width, resampling_method).astype(  # type: ignore [arg-type]
            "bool"
        )
        alpha_mask = self.alpha_mask
        if alpha_mask is not None:
            alpha_mask = resize_array(alpha_mask.copy(), height, width, resampling_method)

        return ImageData(
            numpy.ma.MaskedArray(data, mask=mask),
            assets=self.assets,
            crs=self.crs,
            bounds=self.bounds,
            band_names=self.band_names,
            band_descriptions=self.band_descriptions,
            nodata=self.nodata,
            scales=self.scales,
            offsets=self.offsets,
            metadata=self.metadata,
            dataset_statistics=self.dataset_statistics,
            alpha_mask=alpha_mask,
        )

    def clip(self, bbox: BBox) -> "ImageData":
        """Clip data and mask to a bbox."""
        row_slice, col_slice = windows.from_bounds(
            *bbox, transform=self.transform
        ).toslices()

        return ImageData(
            self.array[:, row_slice, col_slice].copy(),
            assets=self.assets,
            crs=self.crs,
            bounds=bbox,
            band_names=self.band_names,
            band_descriptions=self.band_descriptions,
            nodata=self.nodata,
            scales=self.scales,
            offsets=self.offsets,
            metadata=self.metadata,
            dataset_statistics=self.dataset_statistics,
            alpha_mask=self.alpha_mask[row_slice, col_slice].copy()
            if self.alpha_mask is not None
            else None,
        )

    def post_process(
        self,
        in_range: Sequence[IntervalTuple] | None = None,
        out_dtype: str | numpy.number = "uint8",
        color_formula: str | None = None,
        **kwargs: Any,
    ) -> "ImageData":
        """Post-process image data.

        Args:
            in_range (tuple): input min/max bounds value to rescale from.
            out_dtype (str, optional): output datatype after rescaling. Defaults to `uint8`.
            color_formula (str, optional): color-ops formula (see: https://github.com/vincentsarago/color-ops).
            kwargs (optional): keyword arguments to forward to `rio_tiler.utils.linear_rescale`.

        Returns:
            ImageData: new ImageData object with the updated data.

        Examples:
            >>> img.post_process(in_range=((0, 16000), ))

            >>> img.post_process(color_formula="Gamma RGB 4.1")

        """
        img = ImageData(
            self.array.copy(),
            crs=self.crs,
            bounds=self.bounds,
            assets=self.assets,
            metadata=self.metadata,
            dataset_statistics=self.dataset_statistics,
            alpha_mask=self.alpha_mask.copy() if self.alpha_mask is not None else None,
        )

        if in_range:
            img.rescale(in_range, out_dtype=out_dtype, **kwargs)

        if color_formula:
            img.apply_color_formula(color_formula)

        return img

    def render(
        self,
        add_mask: bool = True,
        img_format: str = "PNG",
        colormap: ColorMapType | None = None,
        **kwargs,
    ) -> bytes:
        """Render data to image blob.

        Args:
            add_mask (bool, optional): add mask to output image. Defaults to `True`.
            img_format (str, optional): output image format. Defaults to `PNG`.
            colormap (dict or sequence, optional): RGBA Color Table dictionary or sequence.
            kwargs (optional): keyword arguments to forward to `rio_tiler.utils.render`.

        Returns:
            bytes: image.

        """
        img_format = img_format.upper()

        if img_format == "GTIFF":
            if "transform" not in kwargs:
                kwargs.update({"transform": self.transform})
            if "crs" not in kwargs and self.crs:
                kwargs.update({"crs": self.crs})

        array = self.array.copy()
        mask = self.mask

        datatype_range = self.dataset_statistics or (dtype_ranges[str(array.dtype)],)

        if not colormap:
            format_dtypes = {
                "PNG": ["uint8", "uint16"],
                "JPEG": ["uint8"],
                "WEBP": ["uint8"],
                "JP2OPENJPEG": ["uint8", "int16", "uint16"],
            }
            valid_dtypes = format_dtypes.get(img_format, [])
            if valid_dtypes and array.dtype not in valid_dtypes:
                warnings.warn(
                    f"Invalid type: `{array.dtype}` for the `{img_format}` driver. Data will be rescaled using min/max type bounds or dataset_statistics.",
                    InvalidDatatypeWarning,
                )
                array = rescale_image(array, in_range=datatype_range)
                if mask is not None:
                    mask = linear_rescale(mask, in_range=dtype_ranges[str(array.dtype)])

        if add_mask:
            return render(
                array.data,
                mask,
                img_format=img_format,
                colormap=colormap,
                **kwargs,
            )

        return render(array.data, img_format=img_format, colormap=colormap, **kwargs)

    def to_raster(self, dst_path: str, *, driver: str = "GTIFF", **kwargs: Any) -> None:
        """Save ImageData array to file."""
        if driver.upper() == "GTIFF":
            if "transform" not in kwargs:
                kwargs.update({"transform": self.transform})
            if "crs" not in kwargs and self.crs:
                kwargs.update({"crs": self.crs})

        write_nodata = self.nodata is not None
        count, height, width = self.array.shape

        output_profile = {
            "dtype": self.array.dtype,
            "count": count if write_nodata else count + 1,
            "height": height,
            "width": width,
            "nodata": self.nodata,
        }
        output_profile.update(kwargs)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=NotGeoreferencedWarning,
                module="rasterio",
            )
            with rasterio.open(dst_path, "w", driver=driver, **output_profile) as dst:
                dst.write(self.data, indexes=list(range(1, count + 1)))

                # Use Mask as an alpha band
                if not write_nodata:
                    if ColorInterp.alpha not in dst.colorinterp:
                        dst.colorinterp = *dst.colorinterp[:-1], ColorInterp.alpha
                    dst.write(self.mask.astype(self.array.dtype), indexes=count + 1)

        return

    def statistics(
        self,
        categorical: bool = False,
        categories: list[float] | None = None,
        percentiles: list[int] | None = None,
        hist_options: dict | None = None,
        coverage: numpy.ndarray | None = None,
    ) -> dict[str, BandStatistics]:
        """Return statistics from ImageData."""
        hist_options = hist_options or {}

        stats = get_array_statistics(
            self.array,
            categorical=categorical,
            categories=categories,
            percentiles=percentiles,
            coverage=coverage,
            **hist_options,
        )

        return {
            f"{self.band_names[ix]}": BandStatistics(
                **stats[ix],
                description=self.band_descriptions[ix],
            )
            for ix in range(len(stats))
        }

    def get_coverage_array(
        self,
        shape: dict,
        shape_crs: CRS = WGS84_CRS,
        cover_scale: int = 10,
    ) -> NDArray[numpy.floating]:
        """Get Coverage array for a Geometry.

        Args:
            shape (Dict): GeoJSON geometry or Feature.
            shape_crs (rasterio.crs.CRS): Coordinates Reference System of shape.
            cover_scale (int, optional):
                Scale used when generating coverage estimates of each
                raster cell by vector feature. Coverage is generated by
                rasterizing the feature at a finer resolution than the raster then using a summation to aggregate
                to the raster resolution and dividing by the square of cover_scale
                to get coverage value for each cell. Increasing cover_scale
                will increase the accuracy of coverage values; three orders
                magnitude finer resolution (cover_scale=1000) is usually enough to
                get coverage estimates with <1% error in individual edge cells coverage
                estimates, though much smaller values (e.g., cover_scale=10) are often
                sufficient (<10% error) and require less memory.

        Returns:
            numpy.array: percent coverage.

        Note: code adapted from https://github.com/perrygeo/python-rasterstats/pull/136 by @sgoodm

        """
        shape = _validate_shape_input(shape)

        if self.crs != shape_crs:
            shape = transform_geom(shape_crs, self.crs, shape)

        cover_array = rasterize(
            [(shape, 1)],
            out_shape=(self.height * cover_scale, self.width * cover_scale),
            transform=self.transform * Affine.scale(1 / cover_scale),
            all_touched=True,
            fill=0,
            dtype="uint8",
        )
        cover_array = cover_array.reshape(
            (self.height, cover_scale, self.width, cover_scale)
        ).astype("float32")

        return cover_array.sum(-1).sum(1) / (cover_scale**2)

    def reproject(
        self,
        dst_crs: CRS,
        resolution: tuple[float, float] | None = None,
        reproject_method: WarpResampling = "nearest",
    ) -> "ImageData":
        """Reproject data and mask."""
        dst_transform, w, h = calculate_default_transform(
            self.crs,
            dst_crs,
            self.width,
            self.height,
            *self.bounds,
            resolution=resolution,
        )
        destination = numpy.ma.masked_array(
            numpy.zeros((self.count, h, w), dtype=self.array.dtype),
        )

        destination, _ = reproject(
            self.array,
            destination,
            src_transform=self.transform,
            src_crs=self.crs,
            dst_transform=dst_transform,
            dst_crs=dst_crs,
            resampling=Resampling[reproject_method],
        )

        alpha_mask = self.alpha_mask
        if self.alpha_mask is not None:
            alpha_mask = numpy.zeros((h, w), dtype=self.alpha_mask.dtype)
            alpha_mask, _ = reproject(
                self.alpha_mask,
                alpha_mask,
                src_transform=self.transform,
                src_crs=self.crs,
                dst_transform=dst_transform,
                dst_crs=dst_crs,
                resampling=Resampling[reproject_method],
            )

        bounds = array_bounds(h, w, dst_transform)

        return ImageData(
            destination,
            assets=self.assets,
            crs=dst_crs,
            bounds=bounds,
            band_names=self.band_names,
            band_descriptions=self.band_descriptions,
            nodata=self.nodata,
            scales=self.scales,
            offsets=self.offsets,
            metadata=self.metadata,
            dataset_statistics=self.dataset_statistics,
            alpha_mask=alpha_mask,
        )



"""rio-tiler.mosaic.methods: Mosaic filling methods."""

import warnings
from enum import Enum, EnumMeta

from .defaults import (
    CountMethod,
    FirstMethod,
    HighestMethod,
    LastBandHighMethod,
    LastBandLowMethod,
    LowestMethod,
    MeanMethod,
    MedianMethod,
    StdevMethod,
)


class _DeprecatedMemberMeta(EnumMeta):
    def __getattr__(cls, name: str):
        if name == "lastbandhight":
            warnings.warn(
                "'lastbandhight' is a typo and will be removed in a future version, "
                "use 'lastbandhigh' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return cls.lastbandhigh
        raise AttributeError(f"'{cls.__name__}' has no attribute '{name}'")

    def __getitem__(cls, name: str):
        if name == "lastbandhight":
            warnings.warn(
                "'lastbandhight' is a typo and will be removed in a future version, "
                "use 'lastbandhigh' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return cls.lastbandhigh
        return super().__getitem__(name)


class PixelSelectionMethod(Enum, metaclass=_DeprecatedMemberMeta):
    """rio-tiler.mosaic pixel selection methods"""

    first = FirstMethod
    highest = HighestMethod
    lowest = LowestMethod
    mean = MeanMethod
    median = MedianMethod
    stdev = StdevMethod
    lastbandlow = LastBandLowMethod
    lastbandhigh = LastBandHighMethod
    count = CountMethod


__all__ = [
    "PixelSelectionMethod",
    "CountMethod",
    "FirstMethod",
    "HighestMethod",
    "LastBandHighMethod",
    "LastBandLowMethod",
    "LowestMethod",
    "MeanMethod",
    "MedianMethod",
    "StdevMethod",
]



"""rio-tiler.io"""

from .base import BaseReader, MultiBandReader, MultiBaseReader
from .rasterio import ImageReader, Reader
from .stac import STACReader
from .xarray import XarrayReader

# Keep Compatibility with <4.0
COGReader = Reader

__all__ = [
    "BaseReader",
    "MultiBandReader",
    "MultiBaseReader",
    "ImageReader",
    "Reader",
    "STACReader",
    "XarrayReader",
    "COGReader",
]



"""rio_tiler.experimental.async_geotiff: rio-tiler Async reader built on top of async-geotiff."""

from __future__ import annotations

import abc
import math
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, cast

import attr
import numpy
from async_geotiff import Window
from async_geotiff.enums import ColorInterp
from morecantile import Tile, TileMatrixSet
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.errors import NotGeoreferencedWarning
from rasterio.features import bounds as featureBounds
from rasterio.features import rasterize
from rasterio.transform import array_bounds, from_bounds
from rasterio.warp import calculate_default_transform, reproject
from rasterio.warp import transform as transform_coords
from rasterio.warp import transform_bounds, transform_geom
from rasterio.windows import from_bounds as window_from_bounds

from rio_tiler.constants import WEB_MERCATOR_TMS, WGS84_CRS
from rio_tiler.errors import (
    ExpressionMixingWarning,
    PointOutsideBounds,
    TileOutsideBounds,
)
from rio_tiler.expression import parse_expression
from rio_tiler.io.base import SpatialMixin
from rio_tiler.models import BandStatistics, ImageData, Info, PointData
from rio_tiler.types import BBox, Indexes, RIOResampling, WarpResampling
from rio_tiler.utils import (
    CRS_to_uri,
    _get_width_height,
    _missing_size,
    _validate_shape_input,
    cast_to_sequence,
)

if TYPE_CHECKING:
    from async_geotiff import GeoTIFF, Overview


@attr.s
class AsyncBaseReader(SpatialMixin, metaclass=abc.ABCMeta):
    """Rio-tiler.io AsyncBaseReader."""

    input: Any = attr.ib()
    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    async def __aenter__(self):
        """Support using with Context Managers."""
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Support using with Context Managers."""
        pass

    @abc.abstractmethod
    async def info(self) -> Info:
        """Return Dataset's info.

        Returns:
            rio_tiler.models.Info: Dataset info.

        """
        ...

    @abc.abstractmethod
    async def statistics(self) -> dict[str, BandStatistics]:
        """Return bands statistics from a dataset.

        Returns:
            dict[str, rio_tiler.models.BandStatistics]: bands statistics.

        """
        ...

    @abc.abstractmethod
    async def tile(self, tile_x: int, tile_y: int, tile_z: int) -> ImageData:
        """Read a Map tile from the Dataset.

        Args:
            tile_x (int): Tile's horizontal index.
            tile_y (int): Tile's vertical index.
            tile_z (int): Tile's zoom level index.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and tile spatial info.

        """
        ...

    @abc.abstractmethod
    async def part(self, bbox: BBox, **kwargs: Any) -> ImageData:
        """Read a Part of a Dataset.

        Args:
            bbox (tuple): Output bounds (left, bottom, right, top) in target crs.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        ...

    @abc.abstractmethod
    async def preview(self) -> ImageData:
        """Read a preview of a Dataset.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        ...

    @abc.abstractmethod
    async def point(self, lon: float, lat: float) -> PointData:
        """Read a value from a Dataset.

        Args:
            lon (float): Longitude.
            lat (float): Latitude.

        Returns:
            list: Pixel value per bands/assets.

        """
        ...

    @abc.abstractmethod
    async def feature(self, shape: dict) -> ImageData:
        """Read a Dataset for a GeoJSON feature.

        Args:
            shape (dict): Valid GeoJSON feature.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        ...


@attr.s
class AsyncReader(AsyncBaseReader):
    """Rio-tiler.io AsyncBaseReader."""

    input: GeoTIFF = attr.ib()

    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    colormap: dict | None = attr.ib(default=None)

    async def __aenter__(self):
        """Support using with Context Managers."""
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Support using with Context Managers."""
        pass

    def __attrs_post_init__(self):
        """Post init."""
        self.bounds = self.input.bounds
        self.crs = CRS.from_user_input(self.input.crs)

        self.transform = self.input.transform
        self.height = self.input.height
        self.width = self.input.width

        if self.colormap is None and self.input.colormap is not None:
            self.colormap = self.input.colormap.as_rasterio()

    @property
    def minzoom(self):
        """Return dataset minzoom."""
        return self._minzoom

    @property
    def maxzoom(self):
        """Return dataset maxzoom."""
        return self._maxzoom

    def _get_overview_level(
        self,
        target_res: float,
        zoom_level_strategy: Literal["AUTO", "LOWER", "UPPER"] = "AUTO",
    ) -> int:
        """Return the overview level corresponding to the resolution."""
        src_res = max(abs(self.input.transform.a), abs(self.input.transform.e))

        available_resolutions = [src_res] + [
            min(abs(ovr.transform.a), abs(ovr.transform.e))
            for ovr in self.input.overviews
        ]

        # Based on aiocogeo:
        # https://github.com/geospatial-jeff/aiocogeo/blob/5a1d32c3f22c883354804168a87abb0a2ea1c328/aiocogeo/partial_reads.py#L113-L147
        percentage = {"AUTO": 50, "LOWER": 100, "UPPER": 0}.get(zoom_level_strategy, 50)

        assert 0 <= percentage <= 100
        # Iterate over zoom levels from lowest/coarsest to highest/finest. If the `target_res` is more than `percentage`
        # percent of the way from the zoom level below to the zoom level above, then upsample the zoom level below, else
        # downsample the zoom level above.
        for i in reversed(range(1, len(available_resolutions))):
            res_current = available_resolutions[i]
            res_higher = available_resolutions[i - 1]
            threshold = res_current - (res_current - res_higher) * (percentage / 100.0)
            if target_res > threshold or target_res == res_current:
                return i

        return 0

    async def info(self) -> Info:
        """Return Dataset's info.

        Returns:
            rio_tiler.models.Info: Dataset info.

        """
        if self.input.nodata is not None:
            nodata_type = "Nodata"
        elif any(c == ColorInterp.ALPHA for c in self.input.colorinterp):
            nodata_type = "Alpha"
        elif self.input._mask_ifd:
            nodata_type = "Mask"
        else:
            nodata_type = "None"

        overviews = [
            math.ceil(self.input.width / ovr.width) for ovr in self.input.overviews
        ]

        meta: dict[str, Any] = {
            "bounds": self.bounds,
            "crs": CRS_to_uri(self.crs) or self.crs.to_wkt(),
            # TODO: get we can band metadata from async-geotiff
            "band_metadata": [(f"b{ix+1}", {}) for ix in range(self.input.count)],
            # TODO: get we can band names from async-geotiff
            "band_descriptions": [
                (f"b{ix+1}", f"b{ix+1}") for ix in range(self.input.count)
            ],
            "dtype": self.input.dtype.name,
            "colorinterp": [ix.name for ix in self.input.colorinterp],
            "nodata_type": nodata_type,
            # additional info (not in default model)
            "driver": "GTiff",
            "count": self.input.count,
            "width": self.input.width,
            "height": self.input.height,
            "overviews": overviews,
            "scales": self.input.scales,
            "offsets": self.input.offsets,
        }

        if self.colormap:
            meta.update({"colormap": self.colormap})

        if nodata_type == "Nodata":
            meta.update({"nodata_value": self.input.nodata})

        return Info.model_validate(meta)

    async def statistics(
        self,
        categorical: bool = False,
        categories: list[float] | None = None,
        percentiles: list[int] | None = None,
        hist_options: dict | None = None,
        max_size: int = 1024,
        indexes: Indexes | None = None,
        expression: str | None = None,
        **kwargs: Any,
    ) -> dict[str, BandStatistics]:
        """Return bands statistics from a dataset.

        Args:
            categorical (bool): treat input data as categorical data. Defaults to False.
            categories (list of numbers, optional): list of categories to return value for.
            percentiles (list of numbers, optional): list of percentile values to calculate. Defaults to `[2, 98]`.
            hist_options (dict, optional): Options to forward to numpy.histogram function.
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio. Defaults to 1024.
            indexes (int or sequence of int, optional): Band indexes.
            expression (str, optional): rio-tiler expression (e.g. b1/b2+b3).
            kwargs (optional): Options to forward to `self._read`.

        Returns:
            dict[str, rio_tiler.models.BandStatistics]: bands statistics.

        """
        data = await self.preview(
            max_size=max_size, indexes=indexes, expression=expression, **kwargs
        )

        return data.statistics(
            categorical=categorical,
            categories=categories,
            percentiles=percentiles,
            hist_options=hist_options,
        )

    async def tile(
        self,
        tile_x: int,
        tile_y: int,
        tile_z: int,
        tilesize: int | None = None,
        indexes: Indexes | None = None,
        expression: str | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read a Map tile from the Dataset.

        Args:
            tile_x (int): Tile's horizontal index.
            tile_y (int): Tile's vertical index.
            tile_z (int): Tile's zoom level index.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and tile spatial info.

        """
        if not self.tile_exists(tile_x, tile_y, tile_z):
            raise TileOutsideBounds(
                f"Tile(x={tile_x}, y={tile_y}, z={tile_z}) is outside bounds"
            )

        matrix = self.tms.matrix(tile_z)
        bbox = cast(
            BBox,
            self.tms.xy_bounds(Tile(x=tile_x, y=tile_y, z=tile_z)),
        )

        return await self.part(
            bbox,
            dst_crs=self.tms.rasterio_crs,
            bounds_crs=self.tms.rasterio_crs,
            height=tilesize or matrix.tileHeight,
            width=tilesize or matrix.tileWidth,
            max_size=None,
            indexes=indexes,
            expression=expression,
            **kwargs,
        )

    async def part(  # noqa: C901
        self,
        bbox: BBox,
        dst_crs: CRS | None = None,
        bounds_crs: CRS = WGS84_CRS,
        indexes: Indexes | None = None,
        expression: str | None = None,
        max_size: int | None = None,
        height: int | None = None,
        width: int | None = None,
        unscale: bool = False,
        resampling_method: RIOResampling = "nearest",
        reproject_method: WarpResampling = "nearest",
        **kwargs: Any,
    ) -> ImageData:
        """Read a Part of a Dataset.

        Args:
            bbox (tuple): Output bounds (left, bottom, right, top) in target crs.
            dst_crs (rasterio.crs.CRS, optional): Overwrite target coordinate reference system. Defaults to bounds_crs.
            bounds_crs (rasterio.crs.CRS, optional): Coordinate reference system of the input bounds. Defaults to WGS84.
            indexes (sequence of int or int, optional): Band indexes.
            expression (str, optional): rio-tiler expression (e.g. b1/b2+b3).
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio. Defaults to 1024.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            unscale (bool, optional): Apply 'scales' and 'offsets' on output data value. Defaults to `False`.
            resampling_method (str, optional): GDAL Resampling method to use when resizing. Defaults to "nearest".
            reproject_method (str, optional): GDAL Resampling method to use when reprojecting. Defaults to "nearest".

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        if indexes and expression:
            warnings.warn(
                "Both expression and indexes passed; expression will overwrite indexes parameter.",
                ExpressionMixingWarning,
            )

        if expression:
            indexes = parse_expression(expression)

        indexes = cast_to_sequence(indexes)

        if max_size and (width or height):
            warnings.warn(
                "'max_size' will be ignored with with 'height' or 'width' set.",
                UserWarning,
            )
            max_size = None

        dst_crs = dst_crs or bounds_crs

        # Transform bbox from bounds_crs → dst_crs
        if bounds_crs != dst_crs:
            bbox = transform_bounds(bounds_crs, dst_crs, *bbox, densify_pts=21)

        # 1. Estimate `max` output dimensions
        if dst_crs != self.crs:
            # Case 1: Reprojection needed
            # - reproject bbox to dataset CRS
            # - compute native output shape in dataset resolution/CRS
            # - compute output shape in output CRS
            dst_bounds = transform_bounds(dst_crs, self.crs, *bbox, densify_pts=21)
            native_src_w = max(
                1, round((dst_bounds[2] - dst_bounds[0]) / abs(self.input.transform.a))
            )
            native_src_h = max(
                1, round((dst_bounds[3] - dst_bounds[1]) / abs(self.input.transform.e))
            )
            _, dst_width, dst_height = calculate_default_transform(
                self.crs, dst_crs, native_src_w, native_src_h, *dst_bounds
            )

        else:
            # Case 2: No reprojection needed
            # - keep output dataset bbox as input bbox
            # - compute output shape in dataset resolution/CRS
            dst_bounds = bbox
            dst_width = max(1, round((bbox[2] - bbox[0]) / abs(self.input.transform.a)))
            dst_height = max(1, round((bbox[3] - bbox[1]) / abs(self.input.transform.e)))

        # Case 1: `max_size` is set,
        # compute output shape based on it,
        # respecting aspect ratio of the output bbox (dst_width/dst_height)
        if max_size:
            height, width = _get_width_height(max_size, dst_height, dst_width)

        # Case 2: One of width/height is missing,
        # compute it but keep the aspect ratio from the max output bbox (dst_width/dst_height)
        elif _missing_size(width, height):
            ratio = dst_height / dst_width
            if width:
                height = math.ceil(width * ratio)
            else:
                width = math.ceil(height / ratio)

        # 2. Output shape (in the output CRS)
        height = cast(int, height or dst_height)
        width = cast(int, width or dst_width)

        # 3. Select IFD based on output resolution in dataset CRS
        if dst_crs != self.crs:
            # Get Transform from output shape and bbox
            transform, _, _ = calculate_default_transform(
                dst_crs, self.crs, width, height, *bbox
            )
        else:
            transform = from_bounds(*bbox, width, height)

        target_res = min(abs(transform.a), abs(transform.e))

        dataset: GeoTIFF | Overview = self.input
        if level := self._get_overview_level(target_res):
            dataset = self.input.overviews[level - 1]

        # 4. Build pixel window, clamped to dataset bounds
        rasterio_win = window_from_bounds(*dst_bounds, transform=dataset.transform)
        row_off = math.floor(rasterio_win.row_off)
        col_off = math.floor(rasterio_win.col_off)
        win_width = math.ceil(rasterio_win.width) + 1
        win_height = math.ceil(rasterio_win.height) + 1

        # TODO: add `minimum_overlap` like in reader.part method
        col_end = min(dataset.width, math.ceil(rasterio_win.col_off + rasterio_win.width))
        row_end = min(
            dataset.height, math.ceil(rasterio_win.row_off + rasterio_win.height)
        )
        if col_off >= col_end or row_off >= row_end:
            raise ValueError("Input BBOX and dataset's bounds do not intersect")

        clipped_col_off = max(0, col_off)
        clipped_row_off = max(0, row_off)
        clipped_col_stop = min(dataset.width, col_off + win_width)
        clipped_row_stop = min(dataset.height, row_off + win_height)
        clipped_width = clipped_col_stop - clipped_col_off
        clipped_height = clipped_row_stop - clipped_row_off

        # 5. Read GeotTIFF/Overview dataset with input window in dataset CRS
        img = await self._read(
            dataset,
            indexes=indexes,
            window=Window(
                col_off=clipped_col_off,
                row_off=clipped_row_off,
                width=clipped_width,
                height=clipped_height,
            ),
            unscale=unscale,
        )

        # 6. Reproject/resample using rasterio.warp.reproject
        img = warp(
            img,
            dst_crs=dst_crs,
            dst_bounds=bbox,
            dst_width=width,
            dst_height=height,
            reproject_method=reproject_method,
        )

        if expression:
            img = img.apply_expression(expression)

        return img

    async def preview(
        self,
        indexes: Indexes | None = None,
        expression: str | None = None,
        dst_crs: CRS | None = None,
        max_size: int = 1024,
        height: int | None = None,
        width: int | None = None,
        unscale: bool = False,
        resampling_method: RIOResampling = "nearest",
        reproject_method: WarpResampling = "nearest",
        **kwargs: Any,
    ) -> ImageData:
        """Read a preview of a Dataset.

        Args:
            indexes (sequence of int or int, optional): Band indexes.
            expression (str, optional): rio-tiler expression (e.g. b1/b2+b3).
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio. Defaults to 1024.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            unscale (bool, optional): Apply 'scales' and 'offsets' on output data value. Defaults to `False`.
            resampling_method (str, optional): GDAL Resampling method to use when resizing. Defaults to "nearest".
            reproject_method (str, optional): GDAL Resampling method to use when reprojecting. Defaults to "nearest".

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        if indexes and expression:
            warnings.warn(
                "Both expression and indexes passed; expression will overwrite indexes parameter.",
                ExpressionMixingWarning,
            )

        if expression:
            indexes = parse_expression(expression)

        indexes = cast_to_sequence(indexes)

        if max_size and (width or height):
            warnings.warn(
                "'max_size' will be ignored with with 'height' or 'width' set.",
                UserWarning,
            )
            max_size = None

        # 1. Determine output shape
        # get height/width of the dataset in the output CRS
        dst_width = self.input.width
        dst_height = self.input.height
        if dst_crs and dst_crs != self.crs:
            # Get shape of the dataset in the output CRS
            _, dst_width, dst_height = calculate_default_transform(
                self.crs, dst_crs, self.width, self.height, *self.bounds
            )

        if max_size:
            height, width = _get_width_height(max_size, dst_height, dst_width)

        elif _missing_size(width, height):
            ratio = dst_height / dst_width
            if width:
                height = math.ceil(width * ratio)
            else:
                width = math.ceil(height / ratio)

        # Shape in the output CRS
        height = height or dst_height
        width = width or dst_width

        # 2. determine overview level to read
        # Output dataset `transform` in the output CRS
        if dst_crs and dst_crs != self.crs:
            proj_bbox = transform_bounds(
                self.crs, dst_crs, *self.input.bounds, densify_pts=21
            )
            transform, _, _ = calculate_default_transform(
                dst_crs,
                self.crs,
                width,
                height,
                *proj_bbox,
            )
        else:
            transform = from_bounds(*self.input.bounds, width, height)

        # 3. Select Overview level
        target_res = min(abs(transform.a), abs(transform.e))
        dataset: GeoTIFF | Overview = self.input
        if level := self._get_overview_level(target_res):
            dataset = self.input.overviews[level - 1]

        # 4. Read data
        img = await self._read(dataset, indexes=indexes, unscale=unscale)
        if expression:
            img = img.apply_expression(expression)

        # 5. Reproject if needed
        if dst_crs and dst_crs != self.crs:
            img = img.reproject(
                dst_crs=dst_crs,
                reproject_method=reproject_method,
            )

        # 6. Resize
        if width != img.width or height != img.height:
            img = img.resize(
                width=width,
                height=height,
                resampling_method=resampling_method,
            )

        return img

    async def point(
        self,
        lon: float,
        lat: float,
        coord_crs: CRS = WGS84_CRS,
        indexes: Indexes | None = None,
        expression: str | None = None,
        unscale: bool = False,
        **kwargs: Any,
    ) -> PointData:
        """Read a value from a Dataset.

        Args:
            lon (float): Longitude.
            lat (float): Latitude.
            coord_crs (rasterio.crs.CRS, optional): Coordinate Reference System of the input coords. Defaults to `epsg:4326`.
            indexes (sequence of int or int, optional): Band indexes.
            expression: (str, optional): Expression to apply on the pixel values. Defaults to `None`.
            unscale: (bool, optional): Apply 'scales' and 'offsets' on output data value. Defaults to `False`.

        Returns:
            PointData: Pixel value per bands/assets.

        """
        if indexes and expression:
            warnings.warn(
                "Both expression and indexes passed; expression will overwrite indexes parameter.",
                ExpressionMixingWarning,
            )

        if expression:
            indexes = parse_expression(expression)

        indexes = cast_to_sequence(indexes)

        coordinates = (lon, lat)
        if coord_crs != self.crs:
            xs, ys = transform_coords(coord_crs, self.crs, [lon], [lat])
            lon, lat = xs[0], ys[0]

        dataset_min_lon, dataset_min_lat, dataset_max_lon, dataset_max_lat = self.bounds
        # check if latitude is inverted
        if dataset_min_lat > dataset_max_lat:
            warnings.warn(
                "BoundingBox of the dataset is inverted (minLat > maxLat).",
                UserWarning,
            )

        dataset_min_lat, dataset_max_lat = (
            min(dataset_min_lat, dataset_max_lat),
            max(dataset_min_lat, dataset_max_lat),
        )
        if not (
            (dataset_min_lon < lon < dataset_max_lon)
            and (dataset_min_lat < lat < dataset_max_lat)
        ):
            raise PointOutsideBounds("Point is outside dataset bounds")

        row, col = self.input.index(lon, lat)
        window = Window(row_off=row, col_off=col, width=1, height=1)
        img = await self._read(
            self.input, indexes=indexes, window=window, unscale=unscale
        )

        pt = PointData(
            img.array[:, 0, 0],
            band_names=img.band_names,
            band_descriptions=img.band_descriptions,
            coordinates=coordinates,
            crs=coord_crs,
            pixel_location=(col, row),
            nodata=img.nodata,
            scales=img.scales,
            offsets=img.offsets,
            metadata=img.metadata,
        )
        if expression:
            pt = pt.apply_expression(expression)

        return pt

    async def feature(
        self,
        shape: dict,
        dst_crs: CRS | None = None,
        shape_crs: CRS = WGS84_CRS,
        indexes: Indexes | None = None,
        expression: str | None = None,
        max_size: int | None = None,
        height: int | None = None,
        width: int | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read a Dataset for a GeoJSON feature.

        Args:
            shape (dict): Valid GeoJSON feature.
            dst_crs (rasterio.crs.CRS, optional): Overwrite target coordinate reference system.
            shape_crs (rasterio.crs.CRS, optional): Input geojson coordinate reference system. Defaults to `epsg:4326`.
            indexes (sequence of int or int, optional): Band indexes.
            expression (str, optional): rio-tiler expression (e.g. b1/b2+b3).
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            **kwargs: Any: Additional parameters to pass to `.part()` method.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        shape = _validate_shape_input(shape)

        if not dst_crs:
            dst_crs = shape_crs

        # Get BBOX of the polygon
        bbox = featureBounds(shape)

        img = await self.part(
            bbox,
            dst_crs=dst_crs,
            bounds_crs=shape_crs,
            indexes=indexes,
            expression=expression,
            max_size=max_size,
            width=width,
            height=height,
            **kwargs,
        )

        if dst_crs != shape_crs:
            shape = transform_geom(shape_crs, dst_crs, shape)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=NotGeoreferencedWarning,
                module="rasterio",
            )
            cutline_mask = rasterize(
                [shape],
                out_shape=(img.height, img.width),
                transform=img.transform,
                all_touched=True,  # Mandatory for matching masks at different resolutions
                default_value=0,
                fill=1,
                dtype="uint8",
            ).astype("bool")

        img.cutline_mask = cutline_mask
        img.array.mask = numpy.where(~cutline_mask, img.array.mask, True)

        return img

    async def _read(
        self,
        dataset: GeoTIFF | Overview,
        indexes: Sequence[int] | None = None,
        window: Window | None = None,
        unscale: bool = False,
    ) -> ImageData:
        """Reader Data from GeoTIFF or Overview."""
        array = await dataset.read(window=window)
        data = array.as_masked()

        if indexes is None:
            indexes = [
                ix
                for ix, c in enumerate(self.input.colorinterp, 1)
                if c != ColorInterp.ALPHA
            ]

        # RGBA datasets
        alpha_mask: numpy.ndarray | None = None
        if ColorInterp.ALPHA in self.input.colorinterp:
            alpha_idx = self.input.colorinterp.index(ColorInterp.ALPHA) + 1
            idx = tuple(indexes) + (alpha_idx,)
            data = data[[ix - 1 for ix in idx]]

            data, alpha_mask = data[:-1], data[-1].data
        else:
            data = data[[ix - 1 for ix in indexes]]

            # if data has Nodata then we simply make sure the mask == the nodata
            if self.input.nodata is not None:
                if numpy.isnan(self.input.nodata):
                    data.mask = numpy.isnan(data.data)
                else:
                    data.mask = data.data == self.input.nodata

        # Handle Scale/Offset
        scales = numpy.array(self.input.scales)[numpy.array(indexes) - 1]
        offsets = numpy.array(self.input.offsets)[numpy.array(indexes) - 1]
        if unscale:
            data = cast(numpy.ma.MaskedArray, data.astype("float32", casting="unsafe"))

            numpy.multiply(data, scales.reshape((-1, 1, 1)), out=data, casting="unsafe")
            numpy.add(data, offsets.reshape((-1, 1, 1)), out=data, casting="unsafe")

            scales = numpy.zeros(len(indexes)) + 1.0
            offsets = numpy.zeros(len(indexes))

        # Get dataset statistics
        stats = []
        if gdal_stats := self.input.stored_stats:
            for ii in indexes:
                if b := gdal_stats.get(ii):
                    if b.min is not None and b.max is not None:
                        stats.append((b.min, b.max))

        # We only add dataset statistics if we have them for all the indexes
        dataset_statistics = stats if len(stats) == len(indexes) else None

        # Create ImageData object
        return ImageData(
            data,
            bounds=array_bounds(array.height, array.width, array.transform),
            crs=CRS.from_user_input(array.crs),
            band_names=[f"b{ix}" for ix in indexes],
            band_descriptions=[f"b{ix}" for ix in indexes],
            dataset_statistics=dataset_statistics,
            nodata=array.nodata,
            alpha_mask=alpha_mask if alpha_mask is not None else None,
            scales=scales.tolist(),
            offsets=offsets.tolist(),
        )


def warp(
    img: ImageData,
    *,
    dst_crs: CRS,
    dst_bounds: BBox,
    dst_width: int,
    dst_height: int,
    reproject_method: WarpResampling = "nearest",
) -> ImageData:
    """Reproject data and mask.

    In the async reader we can't use the rasterio/GDAL WarpedVRT to handle reprojection/resampling,
    so we need to do it manually with `rasterio.warp.reproject`.

    """
    dst_transform = from_bounds(*dst_bounds, dst_width, dst_height)

    destination = numpy.zeros((img.count, dst_height, dst_width), dtype=img.array.dtype)
    destination, _ = reproject(
        img.array,
        destination,
        src_transform=img.transform,
        src_crs=img.crs,
        src_nodata=img.nodata,
        dst_crs=dst_crs,
        dst_transform=dst_transform,
        dst_nodata=img.nodata,
        resampling=Resampling[reproject_method],
    )

    # rasterio doesn't handle masked arrays really well
    # ref: https://github.com/rasterio/rasterio/pull/3531
    mask = ~img.array.mask * numpy.uint8(255)
    alpha_mask = numpy.zeros((img.count, dst_height, dst_width), dtype=numpy.uint8)
    alpha_mask, _ = reproject(
        mask,
        alpha_mask,
        src_transform=img.transform,
        src_crs=img.crs,
        dst_transform=dst_transform,
        dst_crs=dst_crs,
        resampling=Resampling["nearest"],
        src_nodata=0,
        dst_nodata=0,
    )

    return ImageData(
        numpy.ma.MaskedArray(destination, mask=~alpha_mask.astype(bool)),
        assets=img.assets,
        crs=dst_crs,
        bounds=dst_bounds,
        band_names=img.band_names,
        band_descriptions=img.band_descriptions,
        nodata=img.nodata,
        scales=img.scales,
        offsets=img.offsets,
        metadata=img.metadata,
        dataset_statistics=img.dataset_statistics,
    )


# compat
Reader = AsyncReader



"""rio_tiler.io.stac: STAC reader."""

import json
import math
import os
import warnings
from collections.abc import Iterator, Sequence
from threading import Lock
from typing import Any, TypedDict
from urllib.parse import urlparse

import attr
import httpx
import pystac
import rasterio
from affine import Affine
from cachetools import LRUCache, cached
from cachetools.keys import hashkey
from morecantile import TileMatrixSet
from rasterio.crs import CRS
from rasterio.features import bounds as featureBounds
from rasterio.transform import array_bounds, from_bounds

from rio_tiler.constants import WEB_MERCATOR_TMS, WGS84_CRS
from rio_tiler.errors import InvalidAssetName, MissingAssets
from rio_tiler.io.base import BaseReader, MultiBaseReader
from rio_tiler.io.rasterio import Reader
from rio_tiler.types import AssetInfo, AssetType, AssetWithOptions

try:
    from boto3.session import Session as boto3_session

except ImportError:  # pragma: nocover
    boto3_session = None  # type: ignore


lru_cache = LRUCache(maxsize=512)  # type: ignore


DEFAULT_VALID_TYPE = {
    "image/tiff; application=geotiff",
    "image/tiff; application=geotiff; profile=cloud-optimized",
    "image/tiff; profile=cloud-optimized; application=geotiff",
    pystac.MediaType.COG,
    "image/vnd.stac.geotiff; cloud-optimized=true",
    "image/tiff",
    "image/x.geotiff",
    "image/jp2",
    "application/x-hdf5",
    "application/x-hdf",
}

STAC_ALTERNATE_KEY = os.environ.get("RIO_TILER_STAC_ALTERNATE_KEY", None)


def aws_get_object(
    bucket: str,
    key: str,
    request_pays: bool = False,
    client: "boto3_session.client" = None,
) -> bytes:
    """AWS s3 get object content."""
    assert boto3_session is not None, "'boto3' must be installed to use s3:// urls"

    if not client:
        if profile_name := os.environ.get("AWS_PROFILE", None):
            session = boto3_session(profile_name=profile_name)

        else:
            access_key = os.environ.get("AWS_ACCESS_KEY_ID", None)
            secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", None)
            access_token = os.environ.get("AWS_SESSION_TOKEN", None)

            # AWS_REGION is GDAL specific. Later overloaded by standard AWS_DEFAULT_REGION
            region_name = os.environ.get(
                "AWS_DEFAULT_REGION", os.environ.get("AWS_REGION", None)
            )

            session = boto3_session(
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_access_key,
                aws_session_token=access_token,
                region_name=region_name or None,
            )

        # AWS_S3_ENDPOINT and AWS_HTTPS are GDAL config options of vsis3 driver
        # https://gdal.org/user/virtual_file_systems.html#vsis3-aws-s3-files
        endpoint_url = os.environ.get("AWS_S3_ENDPOINT", None)
        if endpoint_url:
            use_https = os.environ.get("AWS_HTTPS", "YES")
            if use_https.upper() in ["YES", "TRUE", "ON"]:
                endpoint_url = "https://" + endpoint_url

            else:
                endpoint_url = "http://" + endpoint_url

        client = session.client("s3", endpoint_url=endpoint_url)

    params = {"Bucket": bucket, "Key": key}
    if request_pays or os.environ.get("AWS_REQUEST_PAYER", "").lower() == "requester":
        params["RequestPayer"] = "requester"

    response = client.get_object(**params)
    return response["Body"].read()


@cached(  # type: ignore
    lru_cache,
    key=lambda filepath, **kargs: hashkey(filepath, json.dumps(kargs)),
    lock=Lock(),
)
def fetch(filepath: str, **kwargs: Any) -> dict:
    """Fetch STAC items.

    A LRU cache is set on top of this function.

    Args:
        filepath (str): STAC item URL.
        kwargs (any): additional options to pass to client.

    Returns:
        dict: STAC Item content.

    """
    parsed = urlparse(filepath)
    if parsed.scheme == "s3":
        bucket = parsed.netloc
        key = parsed.path.strip("/")
        return json.loads(aws_get_object(bucket, key, **kwargs))

    elif parsed.scheme in ["https", "http", "ftp"]:
        resp = httpx.get(filepath, **kwargs)
        resp.raise_for_status()
        return resp.json()

    else:
        with open(filepath, "r") as f:
            return json.load(f)


def _get_assets(
    stac_item: pystac.Item,
    include: set[str] | None = None,
    exclude: set[str] | None = None,
    include_asset_types: set[str] | None = None,
    exclude_asset_types: set[str] | None = None,
) -> Iterator[str]:
    """Get valid asset list.

    Args:
        stac_item (pystac.Item): STAC Item.
        include (Optional[Set[str]]): Only Include specific assets.
        exclude (Optional[Set[str]]): Exclude specific assets.
        include_asset_types (Optional[Set[str]]): Only include some assets base on their type.
        exclude_asset_types (Optional[Set[str]]): Exclude some assets base on their type.

    Yields
        str: valid STAC asset name.

    """
    for asset, asset_info in stac_item.get_assets().items():
        _type = asset_info.media_type

        if exclude and asset in exclude:
            continue

        if (
            _type
            and (exclude_asset_types and _type in exclude_asset_types)
            or (include and asset not in include)
        ):
            continue

        if (
            _type
            and (include_asset_types and _type not in include_asset_types)
            or (include and asset not in include)
        ):
            continue

        yield asset


def _to_pystac_item(item: dict | pystac.Item | None) -> pystac.Item | None:
    """Attr converter to convert to Dict to pystac.Item

    Args:
        item (Union[Dict, pystac.Item]): STAC Item.

    Returns
        pystac.Item: pystac STAC item object.

    """
    if isinstance(item, dict):
        return pystac.Item.from_dict(item)

    return item


class Projection(TypedDict):
    """Projection info extracted from STAC Item."""

    width: int
    height: int
    bounds: tuple[float, float, float, float]
    transform: Affine
    crs: CRS


def _extract_proj_info(
    item: pystac.Item, assets: list[str] | None = None
) -> Projection | None:
    """Extract projection info from STAC Item.

    Args:
        item (pystac.Item): STAC Item.
        assets (list[str]): An optional list of assets to use instead of the item's proj value.

    Returns:
        Projection: projection info extracted from the STAC Item.

    """
    if not item.ext.has("proj"):
        return None

    crs = item.ext.proj.crs_string

    # Asset level projection extension
    assets = assets or list(item.assets)
    asset_proj: dict[str, Any] = {}

    for asset in assets:
        asset_info = item.assets[asset]
        if asset_info.ext.has("proj"):
            if all(
                [
                    asset_info.ext.proj.transform,
                    asset_info.ext.proj.shape,
                ]
            ):
                tr = Affine(*asset_info.ext.proj.transform)
                asset_proj[asset] = {
                    "shape": asset_info.ext.proj.shape,
                    "transform": tr,
                    "bounds": array_bounds(
                        asset_info.ext.proj.shape[0],
                        asset_info.ext.proj.shape[1],
                        tr,
                    ),
                    "crs_string": asset_info.ext.proj.crs_string or crs,
                }

    if asset_proj:
        # 1. check that all assets have the same projection info
        # if multiple CRS if will be too expensive to handle
        crss = [p["crs_string"] for p in asset_proj.values()]
        if crs:
            crss.append(crs)

        if len(set(crss)) == 1:
            crs = next(iter(crss))
            if not crs:
                return None

            # 2. create unified bounds, transform and shape for the item based on the assets info
            bounds = (
                min(p["bounds"][0] for p in asset_proj.values()),
                min(p["bounds"][1] for p in asset_proj.values()),
                max(p["bounds"][2] for p in asset_proj.values()),
                max(p["bounds"][3] for p in asset_proj.values()),
            )

            # 3. Get the highest resolution asset and use its transform to calculate
            # the theoritical shape of the item
            highest_res_asset = min(
                asset_proj.items(),
                key=lambda p: (
                    abs(p[1]["transform"][0]),  # pixel width
                    abs(p[1]["transform"][4]),  # pixel height
                ),
            )
            xres = highest_res_asset[1]["transform"][0]
            yres = highest_res_asset[1]["transform"][4]

            # 4. Get dataset shape from bounds and resolution
            height = abs(math.floor((bounds[3] - bounds[1]) / yres))
            width = abs(math.floor((bounds[2] - bounds[0]) / xres))

            transform = from_bounds(*bounds, width, height)

            return Projection(
                width=width,
                height=height,
                bounds=bounds,
                transform=transform,
                crs=CRS.from_string(crs),
            )

    # Item Level projection extension
    if all(
        [
            item.ext.proj.transform,
            item.ext.proj.shape,
            crs,
        ]
    ):
        height, width = item.ext.proj.shape
        transform = Affine(*item.ext.proj.transform)
        bounds = array_bounds(height, width, transform)

        return Projection(
            width=width,
            height=height,
            bounds=bounds,
            transform=transform,
            crs=CRS.from_string(crs),
        )

    return None


@attr.s
class STACReader(MultiBaseReader):
    """STAC Reader.

    Attributes:
        input (str): STAC Item path, URL or S3 URL.
        item (dict or pystac.Item, STAC): Stac Item.
        tms (morecantile.TileMatrixSet, optional): TileMatrixSet grid definition. Defaults to `WebMercatorQuad`.
        minzoom (int, optional): Set minzoom for the tiles.
        maxzoom (int, optional): Set maxzoom for the tiles.
        include_assets (set of string, optional): Only Include specific assets.
        exclude_assets (set of string, optional): Exclude specific assets.
        include_asset_types (set of string, optional): Only include some assets base on their type.
        exclude_asset_types (set of string, optional): Exclude some assets base on their type.
        default_assets (list of string or dict, optional): Default assets to use if none are defined.
        reader (rio_tiler.io.BaseReader, optional): rio-tiler Reader. Defaults to `rio_tiler.io.Reader`.
        reader_options (dict, optional): Additional option to forward to the Reader. Defaults to `{}`.
        fetch_options (dict, optional): Options to pass to `rio_tiler.io.stac.fetch` function fetching the STAC Items. Defaults to `{}`.

    Examples:
        >>> with STACReader(stac_path) as stac:
            stac.tile(...)

        >>> with STACReader(stac_path, reader=MyCustomReader, reader_options={...}) as stac:
            stac.tile(...)

        >>> my_stac = {
                "type": "Feature",
                "stac_version": "1.0.0",
                ...
            }
            with STACReader(None, item=my_stac) as stac:
                # the dict will be translated to a pystac item
                assert isinstance(stac.item, pystac.Item)
                stac.tile(...)

    """

    input: str | None = attr.ib()
    item: pystac.Item | None = attr.ib(default=None, converter=_to_pystac_item)

    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)
    minzoom: int | None = attr.ib(default=None)
    maxzoom: int | None = attr.ib(default=None)

    include_assets: set[str] | None = attr.ib(default=None)
    exclude_assets: set[str] | None = attr.ib(default=None)

    include_asset_types: set[str] = attr.ib(default=DEFAULT_VALID_TYPE)
    exclude_asset_types: set[str] | None = attr.ib(default=None)

    assets: Sequence[str] = attr.ib(init=False)
    default_assets: Sequence[AssetType] | None = attr.ib(default=None)

    reader: type[BaseReader] = attr.ib(default=Reader)
    reader_options: dict = attr.ib(factory=dict)

    fetch_options: dict = attr.ib(factory=dict)

    ctx: type[rasterio.Env] = attr.ib(default=rasterio.Env)

    def __attrs_post_init__(self):
        """Fetch STAC Item and get list of valid assets."""
        self.item = self.item or pystac.Item.from_dict(
            fetch(self.input, **self.fetch_options), self.input
        )
        self.assets = self.get_asset_list()
        if not self.assets:
            raise MissingAssets("No valid asset found. Asset's media types not supported")

        if proj := _extract_proj_info(self.item, assets=self.assets):
            self.height = proj["height"]
            self.width = proj["width"]
            self.bounds = proj["bounds"]
            self.transform = proj["transform"]
            self.crs = proj["crs"]
        else:
            self.bounds = (
                tuple(self.item.bbox)
                if self.item.bbox
                else featureBounds(self.item.geometry)
            )
            self.crs = WGS84_CRS

        self.minzoom = self.minzoom if self.minzoom is not None else self._minzoom
        self.maxzoom = self.maxzoom if self.maxzoom is not None else self._maxzoom

    def get_asset_list(self) -> list[str]:
        """Get valid asset list"""
        return list(
            _get_assets(
                self.item,
                include=self.include_assets,
                exclude=self.exclude_assets,
                include_asset_types=self.include_asset_types,
                exclude_asset_types=self.exclude_asset_types,
            )
        )

    def _get_options(
        self,
        asset: AssetWithOptions,
        metadata: pystac.Asset,
    ) -> tuple[dict[str, Any], dict[str, Any]]:
        method_options: dict[str, Any] = {}
        reader_options: dict[str, Any] = {}

        # Indexes
        if indexes := asset.get("indexes"):
            method_options["indexes"] = indexes
        # Expression
        if expr := asset.get("expression"):
            method_options["expression"] = expr
        # Bands
        if bands := asset.get("bands"):
            stac_bands = (
                metadata.extra_fields.get("bands")
                or metadata.extra_fields.get("eo:bands")  # V1.0
            )
            if not stac_bands:
                raise ValueError(
                    "Asset does not have 'bands' metadata, unable to use 'bands' option"
                )

            # There is no standard for precedence between 'eo:common_name' and 'name'
            # in STAC specification, so we will use 'eo:common_name' if it exists,
            # otherwise fallback to 'name', and if not exist use the band index as last resource.
            common_to_variable = {
                b.get("eo:common_name") or b.get("common_name") or b.get("name") or ix: ix
                for ix, b in enumerate(stac_bands, 1)
            }
            band_indexes: list[int] = []
            for b in bands:
                if idx := common_to_variable.get(b):
                    band_indexes.append(idx)
                else:
                    raise ValueError(
                        f"Band '{b}' not found in asset metadata, unable to use 'bands' option"
                    )

                method_options["indexes"] = band_indexes

        return reader_options, method_options

    def _get_asset_info(self, asset: AssetType) -> AssetInfo:  # noqa: C901
        """Validate asset names and return asset's info.

        Args:
            asset (AssetType): STAC asset name.

        Returns:
            AssetInfo: STAC asset info.

        """
        if isinstance(asset, str):
            asset = {"name": asset}

        if not asset.get("name"):
            raise ValueError("asset dictionary does not have `name` key")

        asset_name = asset["name"]
        if asset_name not in self.assets:
            raise InvalidAssetName(
                f"'{asset_name}' is not valid, should be one of {self.assets}"
            )

        asset_info = self.item.assets[asset_name]

        reader_options, method_options = self._get_options(asset, asset_info)

        info = AssetInfo(
            url=asset_info.get_absolute_href() or asset_info.href,
            name=asset_name,
            media_type=asset_info.media_type,
            reader_options=reader_options,
            method_options=method_options,
        )

        info["metadata"] = asset_info.extra_fields

        if STAC_ALTERNATE_KEY and asset_info.extra_fields.get("alternate"):
            if alternate := asset_info.extra_fields["alternate"].get(STAC_ALTERNATE_KEY):
                info["url"] = alternate["href"]

        # https://github.com/stac-extensions/file
        if head := asset_info.extra_fields.get("file:header_size"):
            info["env"] = {"GDAL_INGESTED_BYTES_AT_OPEN": head}

        # https://github.com/stac-extensions/raster
        if (
            stac_bands := (
                asset_info.extra_fields.get("bands")
                or asset_info.extra_fields.get("raster:bands")  # V1.0
            )
        ) and "expression" not in method_options:
            stats = [
                (float(b["statistics"]["minimum"]), float(b["statistics"]["maximum"]))
                for b in stac_bands
                if {"minimum", "maximum"}.issubset(b.get("statistics", {}))
            ]
            # check that stats data are all double and raise warning if not
            if (
                stats
                and all(isinstance(v, (int, float)) for stat in stats for v in stat)
                and len(stats) == len(stac_bands)
            ):
                info["dataset_statistics"] = stats
            else:
                warnings.warn(
                    "Some statistics data in STAC are invalid, they will be ignored.",
                    UserWarning,
                )

        return info



"""rio-tiler."""

__version__ = "9.0.4"

from . import (
    colormap,
    constants,
    errors,
    expression,
    io,
    mosaic,
    profiles,
    reader,
    tasks,
    utils,
)

__all__ = [
    "colormap",
    "constants",
    "errors",
    "expression",
    "io",
    "mosaic",
    "profiles",
    "reader",
    "tasks",
    "utils",
]



"""rio_tiler.mosaic: create tile from multiple assets."""

import warnings
from collections.abc import Callable, Sequence
from inspect import isclass
from typing import Any, cast

import numpy
from rasterio.crs import CRS

from rio_tiler.constants import MAX_THREADS
from rio_tiler.errors import (
    EmptyMosaicError,
    InvalidMosaicMethod,
    PointOutsideBounds,
    TileOutsideBounds,
)
from rio_tiler.models import ImageData, PointData
from rio_tiler.mosaic.methods.base import MosaicMethodBase
from rio_tiler.mosaic.methods.defaults import FirstMethod
from rio_tiler.tasks import create_tasks, filter_tasks
from rio_tiler.types import BBox
from rio_tiler.utils import _chunks, resize_array


def mosaic_reader(  # noqa: C901
    mosaic_assets: Sequence,
    reader: Callable[..., ImageData],
    *args: Any,
    pixel_selection: type[MosaicMethodBase] | MosaicMethodBase = FirstMethod,
    chunk_size: int | None = None,
    threads: int = MAX_THREADS,
    allowed_exceptions: tuple = (TileOutsideBounds,),
    **kwargs,
) -> tuple[ImageData, list]:
    """Merge multiple assets.

    Args:

        mosaic_assets (sequence): List of assets.
        reader (callable): Reader function. The function MUST take `(asset, *args, **kwargs)` as arguments, and MUST return an ImageData.
        args (Any): Argument to forward to the reader function.
        pixel_selection (MosaicMethod, optional): Instance of MosaicMethodBase class. Defaults to `rio_tiler.mosaic.methods.defaults.FirstMethod`.
        chunk_size (int, optional): Control the number of asset to process per loop.
        threads (int, optional): Number of threads to use. If <= 1, runs single threaded without an event loop. By default reads from the MAX_THREADS environment variable, and if not found defaults to multiprocessing.cpu_count() * 5.
        allowed_exceptions (tuple, optional): List of exceptions which will be ignored. Note: `TileOutsideBounds` is likely to be raised and should be included in the allowed_exceptions. Defaults to `(TileOutsideBounds, )`.
        kwargs (optional): Reader callable's keywords options.

    Returns:
        tuple: ImageData and assets (list).

    Examples:
        >>> def reader(asset: str, *args, **kwargs) -> ImageData:
                with Reader(asset) as src:
                    return src.tile(*args, **kwargs)

            x, y, z = 10, 10, 4
            img = mosaic_reader(["cog.tif", "cog2.tif"], reader, x, y, z)

        >>> def reader(asset: str, *args, **kwargs) -> ImageData:
                with Reader(asset) as src:
                    return src.preview(*args, **kwargs)

            img = mosaic_reader(["cog.tif", "cog2.tif"], reader)


    """
    if isclass(pixel_selection):
        pixel_selection = cast(type[MosaicMethodBase], pixel_selection)

        if issubclass(pixel_selection, MosaicMethodBase):
            pixel_selection = pixel_selection()

    if not isinstance(pixel_selection, MosaicMethodBase):
        raise InvalidMosaicMethod(
            "Mosaic filling algorithm should be an instance of "
            "'rio_tiler.mosaic.methods.base.MosaicMethodBase'"
        )

    if not chunk_size:
        chunk_size = threads if threads > 1 else len(mosaic_assets)

    assets_used: list = []
    crs: CRS | None
    bounds: BBox | None
    band_names: list[str]

    for chunks in _chunks(mosaic_assets, chunk_size):
        tasks = create_tasks(reader, chunks, threads, *args, **kwargs)
        for img, asset in filter_tasks(
            tasks,
            allowed_exceptions=allowed_exceptions,
        ):
            # On the first Image we set the properties
            if len(assets_used) == 0:
                crs = img.crs
                bounds = img.bounds
                band_names = img.band_names
                band_descriptions = img.band_descriptions
                pixel_selection.cutline_mask = img.cutline_mask
                pixel_selection.width = img.width
                pixel_selection.height = img.height
                pixel_selection.count = img.count

            assert (
                img.count == pixel_selection.count
            ), "Assets HAVE TO have the same number of bands"
            if any(
                [
                    img.width != pixel_selection.width,
                    img.height != pixel_selection.height,
                ]
            ):
                warnings.warn(
                    "Cannot concatenate images with different sizes. Will resize using first asset's width/height.",
                    UserWarning,
                )
                h = pixel_selection.height
                w = pixel_selection.width
                pixel_selection.feed(
                    numpy.ma.MaskedArray(
                        resize_array(img.array.data, h, w),
                        mask=resize_array(img.array.mask * 1, h, w).astype("bool"),
                    )
                )

            else:
                pixel_selection.feed(img.array)

            assets_used.append(asset)

            if pixel_selection.is_done and pixel_selection.data is not None:
                return (
                    ImageData(
                        pixel_selection.data,
                        assets=assets_used,
                        crs=crs,
                        bounds=bounds,
                        band_names=band_names,
                        band_descriptions=band_descriptions,
                        metadata={
                            "mosaic_method": pixel_selection.__class__.__name__,
                            "mosaic_assets_count": len(mosaic_assets),
                            "mosaic_assets_used": len(assets_used),
                        },
                    ),
                    assets_used,
                )

    if pixel_selection.data is None:
        raise EmptyMosaicError("Method returned an empty array")

    return (
        ImageData(
            pixel_selection.data,
            assets=assets_used,
            crs=crs,
            bounds=bounds,
            band_names=band_names,
            band_descriptions=band_descriptions,
            metadata={
                "mosaic_method": pixel_selection.__class__.__name__,
                "mosaic_assets_count": len(mosaic_assets),
                "mosaic_assets_used": len(assets_used),
            },
        ),
        assets_used,
    )


def mosaic_point_reader(
    mosaic_assets: Sequence,
    reader: Callable[..., PointData],
    *args: Any,
    pixel_selection: type[MosaicMethodBase] | MosaicMethodBase = FirstMethod,
    chunk_size: int | None = None,
    threads: int = MAX_THREADS,
    allowed_exceptions: tuple = (PointOutsideBounds,),
    **kwargs,
) -> tuple[PointData, list]:
    """Merge multiple assets.

    Args:

        mosaic_assets (sequence): List of assets.
        reader (callable): Reader function. The function MUST take `(asset, *args, **kwargs)` as arguments, and MUST return a PointData object.
        args (Any): Argument to forward to the reader function.
        pixel_selection (MosaicMethod, optional): Instance of MosaicMethodBase class. Defaults to `rio_tiler.mosaic.methods.defaults.FirstMethod`.
        chunk_size (int, optional): Control the number of asset to process per loop.
        threads (int, optional): Number of threads to use. If <= 1, runs single threaded without an event loop. By default reads from the MAX_THREADS environment variable, and if not found defaults to multiprocessing.cpu_count() * 5.
        allowed_exceptions (tuple, optional): List of exceptions which will be ignored. Note: `PointOutsideBounds` is likely to be raised and should be included in the allowed_exceptions. Defaults to `(TileOutsideBounds, )`.
        kwargs (optional): Reader callable's keywords options.

    Returns:
        tuple: PointData and assets (list).

    Examples:
        >>> def reader(asset: str, *args, **kwargs) -> PointData:
                with Reader(asset) as src:
                    return src.point(*args, **kwargs)

            pt = mosaic_point_reader(["cog.tif", "cog2.tif"], reader, 0, 0)

    """
    if isclass(pixel_selection):
        pixel_selection = cast(type[MosaicMethodBase], pixel_selection)

        if issubclass(pixel_selection, MosaicMethodBase):
            pixel_selection = pixel_selection()

    if not isinstance(pixel_selection, MosaicMethodBase):
        raise InvalidMosaicMethod(
            "Mosaic filling algorithm should be an instance of "
            "'rio_tiler.mosaic.methods.base.MosaicMethodBase'"
        )

    if not chunk_size:
        chunk_size = threads if threads > 1 else len(mosaic_assets)

    assets_used: list = []
    crs: CRS | None
    coordinates: tuple[float, float] | None
    band_names: list[str]

    for chunks in _chunks(mosaic_assets, chunk_size):
        tasks = create_tasks(reader, chunks, threads, *args, **kwargs)
        for pt, asset in filter_tasks(
            tasks,
            allowed_exceptions=allowed_exceptions,
        ):
            # On the first Image we set the properties
            if len(assets_used) == 0:
                crs = pt.crs
                coordinates = pt.coordinates
                band_names = pt.band_names
                band_descriptions = pt.band_descriptions
                pixel_selection.width = 1
                pixel_selection.height = 1
                pixel_selection.count = pt.count

            assert (
                pt.count == pixel_selection.count
            ), "Assets HAVE TO have the same number of bands"

            assets_used.append(asset)
            pixel_selection.feed(pt.array)

            if pixel_selection.is_done and pixel_selection.data is not None:
                return (
                    PointData(
                        pixel_selection.data,
                        assets=assets_used,
                        crs=crs,
                        coordinates=coordinates,
                        band_names=band_names,
                        band_descriptions=band_descriptions,
                        metadata={
                            "mosaic_method": pixel_selection.__class__.__name__,
                            "mosaic_assets_count": len(mosaic_assets),
                            "mosaic_assets_used": len(assets_used),
                        },
                    ),
                    assets_used,
                )

    if pixel_selection.data is None:
        raise EmptyMosaicError("Method returned an empty array")

    return (
        PointData(
            pixel_selection.data,
            assets=assets_used,
            crs=crs,
            coordinates=coordinates,
            band_names=band_names,
            band_descriptions=band_descriptions,
            metadata={
                "mosaic_method": pixel_selection.__class__.__name__,
                "mosaic_assets_count": len(mosaic_assets),
                "mosaic_assets_used": len(assets_used),
            },
        ),
        assets_used,
    )



"""rio_tiler.io.xarray: Xarray Reader."""

from __future__ import annotations

import math
import os
import warnings
from typing import Any, TypedDict, cast

import attr
import numpy
from morecantile import Tile, TileMatrixSet
from rasterio import windows
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.errors import NotGeoreferencedWarning
from rasterio.features import bounds as featureBounds
from rasterio.features import rasterize
from rasterio.transform import from_bounds, rowcol
from rasterio.warp import calculate_default_transform
from rasterio.warp import transform as transform_coords
from rasterio.warp import transform_bounds, transform_geom

from rio_tiler.constants import WEB_MERCATOR_CRS, WEB_MERCATOR_TMS, WGS84_CRS
from rio_tiler.errors import (
    InvalidGeographicBounds,
    MaxArraySizeError,
    MissingCRS,
    PointOutsideBounds,
    TileOutsideBounds,
)
from rio_tiler.io.base import BaseReader
from rio_tiler.models import BandStatistics, ImageData, Info, PointData
from rio_tiler.types import BBox, Indexes, NoData, RIOResampling, WarpResampling
from rio_tiler.utils import (
    CRS_to_uri,
    _get_width_height,
    _missing_size,
    _validate_shape_input,
    cast_to_sequence,
    get_array_statistics,
)

try:
    import xarray
except ImportError:  # pragma: nocover
    xarray = None  # type: ignore

try:
    import rioxarray
except ImportError:  # pragma: nocover
    rioxarray = None  # type: ignore


MAX_ARRAY_SIZE = int(os.environ.get("RIO_TILER_MAX_ARRAY_SIZE", 1_000_000_000))  # 1Gb


class Options(TypedDict, total=False):
    """Reader Options."""

    nodata: NoData | None


@attr.s
class XarrayReader(BaseReader):
    """Xarray Reader.

    Attributes:
        dataset (xarray.DataArray): Xarray DataArray dataset.
        tms (morecantile.TileMatrixSet, optional): TileMatrixSet grid definition. Defaults to `WebMercatorQuad`.

    Examples:
        >>> ds = xarray.open_dataset(
                "https://pangeo.blob.core.windows.net/pangeo-public/daymet-rio-tiler/na-wgs84.zarr",
                engine="zarr",
                decode_coords="all",
                consolidated=True,
            )
            da = ds["tmax"]

            with XarrayReader(da) as dst:
                img = dst.tile(...)

    """

    input: xarray.DataArray = attr.ib()

    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    options: Options = attr.ib()

    _dims: list = attr.ib(init=False, factory=list)

    @options.default
    def _options_default(self):
        return {}

    def __attrs_post_init__(self):
        """Set bounds and CRS."""
        assert xarray is not None, "xarray must be installed to use XarrayReader"
        assert rioxarray is not None, "rioxarray must be installed to use XarrayReader"

        # NOTE: rioxarray returns **ordered** bounds in form of (minx, miny, maxx, maxx)
        self.bounds = tuple(self.input.rio.bounds())
        self.crs = self.input.rio.crs
        if not self.crs:
            raise MissingCRS(
                "Dataset doesn't have CRS information, please add it before using rio-tiler (e.g. `ds.rio.write_crs('epsg:4326', inplace=True)`)"
            )

        # adds half x/y resolution on each values
        # https://github.com/corteva/rioxarray/issues/645#issuecomment-1461070634
        xres, yres = map(abs, self.input.rio.resolution())
        if self.crs == WGS84_CRS and (
            self.bounds[0] + xres / 2 < -180
            or self.bounds[1] + yres / 2 < -90
            or self.bounds[2] - xres / 2 > 180
            or self.bounds[3] - yres / 2 > 90
        ):
            raise InvalidGeographicBounds(
                f"Invalid geographic bounds: {self.bounds}. Must be within (-180, -90, 180, 90)."
            )

        self.transform = self.input.rio.transform()
        self.height = self.input.rio.height
        self.width = self.input.rio.width

        self._dims = [
            d
            for d in self.input.dims
            if d
            not in [
                self.input.rio.x_dim,
                self.input.rio.y_dim,
                "spatial_ref",
                "crs_wkt",
                "grid_mapping",
            ]
        ]
        assert len(self._dims) in [0, 1], "Can't handle >=4D DataArray"

    @property
    def minzoom(self):
        """Return dataset minzoom."""
        return self._minzoom

    @property
    def maxzoom(self):
        """Return dataset maxzoom."""
        return self._maxzoom

    @property
    def band_descriptions(self) -> list[str]:
        """
        Return list of `band descriptions` in DataArray.

        `Bands` are all dimensions not defined as spatial dims by rioxarray.
        """
        if not self._dims:
            coords_name = [
                d
                for d in self.input.coords
                if d
                not in [
                    self.input.rio.x_dim,
                    self.input.rio.y_dim,
                    "spatial_ref",
                    "crs_wkt",
                    "grid_mapping",
                ]
            ]
            if coords_name:
                return [str(self.input.coords[coords_name[0]].data)]

            return [self.input.name or "b1"]  # type: ignore

        return [str(band) for d in self._dims for band in self.input[d].values]

    def info(self) -> Info:
        """Return xarray.DataArray info."""
        metadata = [band.attrs for d in self._dims for band in self.input[d]] or [{}]

        nodata_type = "None"
        if self.options.get("nodata", self.input.rio.nodata) is not None:
            nodata_type = "Nodata"

        meta = {
            "bounds": self.bounds,
            "crs": CRS_to_uri(self.crs) or self.crs.to_wkt(),
            "band_metadata": [(f"b{ix}", v) for ix, v in enumerate(metadata, 1)],
            "band_descriptions": [
                (f"b{ix}", v or f"b{ix}")
                for ix, v in enumerate(self.band_descriptions, 1)
            ],
            "dtype": str(self.input.dtype),
            "nodata_type": nodata_type,
            "name": self.input.name,
            "count": self.input.rio.count,
            "width": self.input.rio.width,
            "height": self.input.rio.height,
            "dimensions": self.input.dims,
            "attrs": {
                k: (v.tolist() if isinstance(v, (numpy.ndarray, numpy.generic)) else v)
                for k, v in self.input.attrs.items()
            },
        }

        return Info(**meta)

    def _sel_indexes(
        self,
        indexes: Indexes | None = None,
    ) -> tuple[xarray.DataArray, list[str]]:
        """Select `band` indexes in DataArray."""
        da = self.input
        band_descriptions = self.band_descriptions

        if indexes := cast_to_sequence(indexes):
            assert all(v > 0 for v in indexes), "Indexes value must be >= 1"
            if da.ndim == 2:
                if set(indexes) != set({1}):
                    raise ValueError(
                        f"Invalid indexes {indexes} for array of shape {da.shape}"
                    )

                return da, band_descriptions

            indexes = [idx - 1 for idx in indexes]

            da = da[indexes]
            band_descriptions = [band_descriptions[idx] for idx in indexes]

        return da, band_descriptions

    def statistics(
        self,
        categorical: bool = False,
        categories: list[float] | None = None,
        percentiles: list[int] | None = None,
        hist_options: dict | None = None,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        **kwargs: Any,
    ) -> dict[str, BandStatistics]:
        """Return statistics from a dataset."""
        hist_options = hist_options or {}

        da, band_descriptions = self._sel_indexes(indexes)

        if nodata is not None:
            da = da.rio.write_nodata(nodata)
        elif (nodata := self.options.get("nodata")) is not None:
            da = da.rio.write_nodata(nodata)

        data = da.to_masked_array()
        data.mask |= data.data == da.rio.nodata

        stats = get_array_statistics(
            data,
            categorical=categorical,
            categories=categories,
            percentiles=percentiles,
            **hist_options,
        )

        return {
            f"b{ix +1}": BandStatistics(
                **val,
                description=band_descriptions[ix],
            )
            for ix, val in enumerate(stats)
        }

    def tile(
        self,
        tile_x: int,
        tile_y: int,
        tile_z: int,
        tilesize: int | None = None,
        reproject_method: WarpResampling = "nearest",
        auto_expand: bool = True,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        out_dtype: str | numpy.dtype | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read a Web Map tile from a dataset.

        Args:
            tile_x (int): Tile's horizontal index.
            tile_y (int): Tile's vertical index.
            tile_z (int): Tile's zoom level index.
            tilesize (int, optional): Output image size. Defaults to `256`.
            reproject_method (WarpResampling, optional): WarpKernel resampling algorithm. Defaults to `nearest`.
            auto_expand (boolean, optional): When True, rioxarray's clip_box will expand clip search if only 1D raster found with clip. When False, will throw `OneDimensionalRaster` error if only 1 x or y data point is found. Defaults to True.
            nodata (int or float, optional): Overwrite dataset internal nodata value.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and tile spatial info.

        """
        if not self.tile_exists(tile_x, tile_y, tile_z):
            raise TileOutsideBounds(
                f"Tile(x={tile_x}, y={tile_y}, z={tile_z}) is outside bounds"
            )

        matrix = self.tms.matrix(tile_z)
        bbox = cast(
            BBox,
            self.tms.xy_bounds(Tile(x=tile_x, y=tile_y, z=tile_z)),
        )

        return self.part(
            bbox,
            dst_crs=self.tms.rasterio_crs,
            bounds_crs=self.tms.rasterio_crs,
            reproject_method=reproject_method,
            auto_expand=auto_expand,
            nodata=nodata,
            indexes=indexes,
            height=tilesize or matrix.tileHeight,
            width=tilesize or matrix.tileWidth,
            out_dtype=out_dtype,
            **kwargs,
        )

    def part(  # noqa: C901
        self,
        bbox: BBox,
        dst_crs: CRS | None = None,
        bounds_crs: CRS = WGS84_CRS,
        reproject_method: WarpResampling = "nearest",
        auto_expand: bool = True,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        max_size: int | None = None,
        height: int | None = None,
        width: int | None = None,
        resampling_method: RIOResampling = "nearest",
        out_dtype: str | numpy.dtype | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read part of a dataset.

        Args:
            bbox (tuple): Output bounds (left, bottom, right, top) in target crs ("dst_crs").
            dst_crs (rasterio.crs.CRS, optional): Overwrite target coordinate reference system.
            bounds_crs (rasterio.crs.CRS, optional): Bounds Coordinate Reference System. Defaults to `epsg:4326`.
            reproject_method (WarpResampling, optional): WarpKernel resampling algorithm. Defaults to `nearest`.
            auto_expand (boolean, optional): When True, rioxarray's clip_box will expand clip search if only 1D raster found with clip. When False, will throw `OneDimensionalRaster` error if only 1 x or y data point is found. Defaults to True.
            nodata (int or float, optional): Overwrite dataset internal nodata value.
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            resampling_method (RIOResampling, optional): RasterIO resampling algorithm. Defaults to `nearest`.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        if max_size and (width or height):
            warnings.warn(
                "'max_size' will be ignored with with 'height' or 'width' set.",
                UserWarning,
            )
            max_size = None

        dst_crs = dst_crs or bounds_crs

        da, band_descriptions = self._sel_indexes(indexes)

        if nodata is not None:
            da = da.rio.write_nodata(nodata)
        elif (nodata := self.options.get("nodata")) is not None:
            da = da.rio.write_nodata(nodata)

        # Forward valid_min/valid_max to the ImageData object
        minv, maxv = da.attrs.get("valid_min"), da.attrs.get("valid_max")
        stats = None
        if minv is not None and maxv is not None:
            stats = ((minv, maxv),) * da.rio.count

        da = da.rio.clip_box(
            *bbox,
            crs=bounds_crs,
            auto_expand=auto_expand,
        )

        if da.nbytes > MAX_ARRAY_SIZE:
            raise MaxArraySizeError(
                f"Maximum array limit {MAX_ARRAY_SIZE} reached, trying to put DataArray of {da.shape} in memory."
            )

        src_width = da.rio.width
        src_height = da.rio.height
        src_bounds = list(da.rio.bounds())
        src_transform = da.rio.transform()

        # Fix for https://github.com/cogeotiff/rio-tiler/issues/654
        #
        # When using `calculate_default_transform` with dataset
        # which span at high/low latitude outside the area_of_use
        # of the WebMercator projection, we `crop` the dataset
        # to get the transform (resolution).
        #
        # Note: Should be handled in gdal 3.8 directly
        # https://github.com/OSGeo/gdal/pull/8775
        if (
            self.crs == WGS84_CRS
            and dst_crs == WEB_MERCATOR_CRS
            and (src_bounds[1] < -85.06 or src_bounds[3] > 85.06)
        ):
            warnings.warn(
                "Adjusting dataset latitudes to avoid re-projection overflow",
                UserWarning,
            )
            src_bounds[1] = max(src_bounds[1], -85.06)
            src_bounds[3] = min(src_bounds[3], 85.06)

            # North->South
            if src_transform.e > 0:
                src_bounds = [src_bounds[0], src_bounds[3], src_bounds[2], src_bounds[1]]
            # West->East
            if src_transform.a < 0:
                src_bounds = [src_bounds[2], src_bounds[1], src_bounds[1], src_bounds[3]]

            w = windows.from_bounds(*src_bounds, transform=src_transform)
            src_height = round(w.height)
            src_width = round(w.width)

        if dst_crs != self.crs:
            # transform of the reprojected dataset
            dst_transform, _, _ = calculate_default_transform(
                self.crs, dst_crs, src_width, src_height, *src_bounds
            )
        else:
            dst_transform = from_bounds(*src_bounds, src_width, src_height)

        if bounds_crs and bounds_crs != dst_crs:
            bbox = transform_bounds(bounds_crs, dst_crs, *bbox, densify_pts=21)

        w, s, e, n = bbox
        # max size of the output dataset for the bbox
        dst_width = max(1, round((e - w) / dst_transform.a))
        dst_height = max(1, round((s - n) / dst_transform.e))

        if max_size:
            height, width = _get_width_height(max_size, dst_height, dst_width)

        elif _missing_size(height, width):
            ratio = dst_height / dst_width
            if width:
                height = math.ceil(width * ratio)
            else:
                width = math.ceil(height / ratio)

        height = height or dst_height
        width = width or dst_width
        da = da.rio.reproject(
            dst_crs,
            shape=(height, width),
            transform=from_bounds(w, s, e, n, width, height),
            resampling=Resampling[reproject_method],
            nodata=nodata,
        )

        arr = da.to_masked_array()
        arr.mask |= arr.data == da.rio.nodata
        if out_dtype:
            arr = arr.astype(out_dtype)

        output_bounds = da.rio._unordered_bounds()
        if output_bounds[1] > output_bounds[3] and da.rio.transform().e > 0:
            yaxis = self.input.dims.index(self.input.rio.y_dim)
            arr = numpy.flip(arr, axis=yaxis)

        return ImageData(
            arr,
            bounds=bbox,
            crs=da.rio.crs,
            dataset_statistics=stats,
            band_descriptions=band_descriptions,
            nodata=da.rio.nodata,
        )

    def preview(  # noqa: C901
        self,
        max_size: int = 1024,
        height: int | None = None,
        width: int | None = None,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        dst_crs: CRS | None = None,
        reproject_method: WarpResampling = "nearest",
        resampling_method: RIOResampling = "nearest",
        out_dtype: str | numpy.dtype | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Return a preview of a dataset.

        Args:
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio. Defaults to 1024.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            nodata (int or float, optional): Overwrite dataset internal nodata value.
            dst_crs (rasterio.crs.CRS, optional): target coordinate reference system.
            reproject_method (WarpResampling, optional): WarpKernel resampling algorithm. Defaults to `nearest`.
            resampling_method (RIOResampling, optional): RasterIO resampling algorithm. Defaults to `nearest`.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        if max_size and (width or height):
            warnings.warn(
                "'max_size' will be ignored with with 'height' or 'width' set.",
                UserWarning,
            )
            max_size = None

        da, band_descriptions = self._sel_indexes(indexes)

        if da.nbytes > MAX_ARRAY_SIZE:
            raise MaxArraySizeError(
                f"Maximum array limit {MAX_ARRAY_SIZE} reached, trying to put DataArray of {da.shape} in memory."
            )

        if nodata is not None:
            da = da.rio.write_nodata(nodata)
        elif (nodata := self.options.get("nodata")) is not None:
            da = da.rio.write_nodata(nodata)

        if dst_crs and dst_crs != self.crs:
            src_width = da.rio.width
            src_height = da.rio.height
            src_bounds = list(da.rio.bounds())
            src_transform = da.rio.transform()

            # Fix for https://github.com/cogeotiff/rio-tiler/issues/654
            #
            # When using `calculate_default_transform` with dataset
            # which span at high/low latitude outside the area_of_use
            # of the WebMercator projection, we `crop` the dataset
            # to get the transform (resolution).
            #
            # Note: Should be handled in gdal 3.8 directly
            # https://github.com/OSGeo/gdal/pull/8775
            if (
                self.crs == WGS84_CRS
                and dst_crs == WEB_MERCATOR_CRS
                and (src_bounds[1] < -85.06 or src_bounds[3] > 85.06)
            ):
                warnings.warn(
                    "Adjusting dataset latitudes to avoid re-projection overflow",
                    UserWarning,
                )
                src_bounds[1] = max(src_bounds[1], -85.06)
                src_bounds[3] = min(src_bounds[3], 85.06)

                # North->South
                if src_transform.e > 0:
                    src_bounds = [
                        src_bounds[0],
                        src_bounds[3],
                        src_bounds[2],
                        src_bounds[1],
                    ]
                # West->East
                if src_transform.a < 0:
                    src_bounds = [
                        src_bounds[2],
                        src_bounds[1],
                        src_bounds[1],
                        src_bounds[3],
                    ]

                w = windows.from_bounds(*src_bounds, transform=src_transform)
                src_height = round(w.height)
                src_width = round(w.width)

            dst_transform, w, h = calculate_default_transform(
                self.crs, dst_crs, src_width, src_height, *src_bounds
            )
            da = da.rio.reproject(
                dst_crs,
                shape=(h, w),
                transform=dst_transform,
                resampling=Resampling[reproject_method],
                nodata=nodata,
            )

        # Forward valid_min/valid_max to the ImageData object
        minv, maxv = da.attrs.get("valid_min"), da.attrs.get("valid_max")
        stats = None
        if minv is not None and maxv is not None:
            stats = ((minv, maxv),) * da.rio.count

        arr = da.to_masked_array()
        if out_dtype:
            arr = cast(numpy.ma.MaskedArray, arr.astype(out_dtype))
        arr.mask |= arr.data == da.rio.nodata

        output_bounds = da.rio._unordered_bounds()
        if output_bounds[1] > output_bounds[3] and da.rio.transform().e > 0:
            yaxis = self.input.dims.index(self.input.rio.y_dim)
            arr = cast(numpy.ma.MaskedArray, numpy.flip(arr, axis=yaxis))

        img = ImageData(
            arr,
            bounds=da.rio.bounds(),
            crs=da.rio.crs,
            dataset_statistics=stats,
            band_descriptions=band_descriptions,
            nodata=da.rio.nodata,
        )

        if max_size:
            height, width = _get_width_height(max_size, img.height, img.width)

        elif _missing_size(height, width):
            ratio = img.height / img.width
            if width:
                height = math.ceil(width * ratio)
            else:
                width = math.ceil(height / ratio)

        if (height and width) and (height != da.rio.height or width != da.rio.width):
            img = img.resize(height, width, resampling_method=resampling_method)

        return img

    def point(
        self,
        lon: float,
        lat: float,
        coord_crs: CRS = WGS84_CRS,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        out_dtype: str | numpy.dtype | None = None,
        **kwargs: Any,
    ) -> PointData:
        """Read a pixel value from a dataset.

        Args:
            lon (float): Longitude.
            lat (float): Latitude.
            coord_crs (rasterio.crs.CRS, optional): Coordinate Reference System of the input coords. Defaults to `epsg:4326`.
            nodata (int or float, optional): Overwrite dataset internal nodata value.

        Returns:
            PointData

        """
        da_lon, da_lat = transform_coords(coord_crs, self.crs, [lon], [lat])

        if not (
            (self.bounds[0] < da_lon[0] < self.bounds[2])
            and (self.bounds[1] < da_lat[0] < self.bounds[3])
        ):
            raise PointOutsideBounds("Point is outside dataset bounds")

        da, band_descriptions = self._sel_indexes(indexes)

        if nodata is not None:
            da = da.rio.write_nodata(nodata)
        elif (nodata := self.options.get("nodata")) is not None:
            da = da.rio.write_nodata(nodata)

        y, x = rowcol(da.rio.transform(), da_lon, da_lat)

        if da.ndim == 2:
            arr = numpy.expand_dims(da[int(y[0]), int(x[0])].to_masked_array(), axis=0)
        else:
            arr = da[:, int(y[0]), int(x[0])].to_masked_array()

        if out_dtype:
            arr = arr.astype(out_dtype)
        arr.mask |= arr.data == da.rio.nodata  # type: ignore [attr-defined]

        return PointData(
            arr,
            coordinates=(lon, lat),
            crs=coord_crs,
            band_descriptions=band_descriptions,
            pixel_location=(x, y),
            nodata=da.rio.nodata,
        )

    def feature(
        self,
        shape: dict,
        dst_crs: CRS | None = None,
        shape_crs: CRS = WGS84_CRS,
        reproject_method: WarpResampling = "nearest",
        auto_expand: bool = True,
        nodata: NoData | None = None,
        indexes: Indexes | None = None,
        max_size: int | None = None,
        height: int | None = None,
        width: int | None = None,
        resampling_method: RIOResampling = "nearest",
        out_dtype: str | numpy.dtype | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read part of a dataset defined by a geojson feature.

        Args:
            shape (dict): Valid GeoJSON feature.
            dst_crs (rasterio.crs.CRS, optional): Overwrite target coordinate reference system.
            shape_crs (rasterio.crs.CRS, optional): Input geojson coordinate reference system. Defaults to `epsg:4326`.
            reproject_method (WarpResampling, optional): WarpKernel resampling algorithm. Defaults to `nearest`.
            auto_expand (boolean, optional): When True, rioxarray's clip_box will expand clip search if only 1D raster found with clip. When False, will throw `OneDimensionalRaster` error if only 1 x or y data point is found. Defaults to True.
            nodata (int or float, optional): Overwrite dataset internal nodata value.
            max_size (int, optional): Limit the size of the longest dimension of the dataset read, respecting bounds X/Y aspect ratio.
            height (int, optional): Output height of the array.
            width (int, optional): Output width of the array.
            resampling_method (RIOResampling, optional): RasterIO resampling algorithm. Defaults to `nearest`.

        Returns:
            rio_tiler.models.ImageData: ImageData instance with data, mask and input spatial info.

        """
        shape = _validate_shape_input(shape)

        if not dst_crs:
            dst_crs = shape_crs

        # Get BBOX of the polygon
        bbox = featureBounds(shape)

        img = self.part(
            bbox,
            dst_crs=dst_crs,
            bounds_crs=shape_crs,
            nodata=nodata,
            indexes=indexes,
            max_size=max_size,
            width=width,
            height=height,
            reproject_method=reproject_method,
            resampling_method=resampling_method,
            out_dtype=out_dtype,
            auto_expand=auto_expand,
            **kwargs,
        )

        if dst_crs != shape_crs:
            shape = transform_geom(shape_crs, dst_crs, shape)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                category=NotGeoreferencedWarning,
                module="rasterio",
            )
            cutline_mask = rasterize(
                [shape],
                out_shape=(img.height, img.width),
                transform=img.transform,
                all_touched=True,  # Mandatory for matching masks at different resolutions
                default_value=0,
                fill=1,
                dtype="uint8",
            ).astype("bool")

        img.cutline_mask = cutline_mask
        img.array.mask = numpy.where(~cutline_mask, img.array.mask, True)

        return img



"""rio_tiler experimental ZarrReaders."""

from __future__ import annotations

import contextlib
from collections.abc import Callable
from functools import cache
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urlparse

import attr
from morecantile import TileMatrixSet
from rasterio.crs import CRS

from rio_tiler.constants import WEB_MERCATOR_TMS, WGS84_CRS
from rio_tiler.errors import InvalidGeographicBounds
from rio_tiler.io.base import BaseReader
from rio_tiler.io.xarray import XarrayReader
from rio_tiler.models import BandStatistics, ImageData, Info, PointData
from rio_tiler.types import BBox

try:
    import obstore
    from zarr.storage import ObjectStore
except ImportError:  # pragma: nocover
    ObjectStore = None  # type: ignore
    obstore = None  # type: ignore

try:
    import rioxarray
    import xarray
except ImportError:  # pragma: nocover
    xarray = None  # type: ignore
    rioxarray = None  # type: ignore

sel_methods = Literal["nearest", "pad", "ffill", "backfill", "bfill"]


@cache
def open_dataset(src_path: str, **kwargs: Any) -> xarray.Dataset:
    """Open Xarray dataset

    Args:
        src_path (str): dataset path.

    Returns:
        xarray.DataTree

    """
    parsed = urlparse(src_path)
    if not parsed.scheme:
        src_path = str(Path(src_path).resolve())
        src_path = "file://" + src_path
    store = obstore.store.from_url(src_path, **kwargs)
    zarr_store = ObjectStore(store=store, read_only=True)
    ds = xarray.open_dataset(
        zarr_store,  # type: ignore
        decode_times=True,
        decode_coords="all",
        consolidated=True,
        engine="zarr",
    )
    return ds


@attr.s
class ZarrReader(BaseReader):
    """Zarr dataset Reader.

    Attributes:
        input (str): dataset path.
        dataset (xarray.Dataset): Xarray dataset.
        tms (morecantile.TileMatrixSet, optional): TileMatrixSet grid definition. Defaults to `WebMercatorQuad`.
        opener (Callable): Xarray dataset opener. Defaults to `open_dataset`.
        opener_options (dict): Options to forward to the opener callable.

    Examples:
        >>> with ZarrReader(
                "s3://mur-sst/zarr-v1",
                opener_options={
                    "skip_signature": True,
                    "region": "us-west-2",
                }
            ) as src:
                print(src)
                print(src.variables)
                img = src.tile(x, y, z, tmax)

    """

    input: str | None = attr.ib()
    dataset: xarray.Dataset | None = attr.ib(default=None)

    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    opener: Callable[..., xarray.Dataset] = attr.ib(default=open_dataset)
    opener_options: dict = attr.ib(factory=dict)
    _ctx_stack: contextlib.ExitStack = attr.ib(init=False, factory=contextlib.ExitStack)

    def __attrs_post_init__(self):
        """Set bounds and CRS."""
        assert xarray is not None, "xarray must be installed to use XarrayReader"
        assert rioxarray is not None, "rioxarray must be installed to use XarrayReader"

        if not self.dataset:
            self.dataset = self._ctx_stack.enter_context(
                self.opener(self.input, **self.opener_options)
            )

        # NOTE: rioxarray returns **ordered** bounds in form of (minx, miny, maxx, maxx)
        self.bounds = tuple(self.dataset.rio.bounds())
        # Make sure we have a valid CRS
        self.dataset = self.dataset.rio.write_crs(
            self.dataset.rio.crs or "epsg:4326",
        )

        self.crs = self.dataset.rio.crs

        # adds half x/y resolution on each values
        # https://github.com/corteva/rioxarray/issues/645#issuecomment-1461070634
        xres, yres = map(abs, self.dataset.rio.resolution())
        if self.crs == WGS84_CRS and (
            self.bounds[0] + xres / 2 < -180
            or self.bounds[1] + yres / 2 < -90
            or self.bounds[2] - xres / 2 > 180
            or self.bounds[3] - yres / 2 > 90
        ):
            raise InvalidGeographicBounds(
                f"Invalid geographic bounds: {self.bounds}. Must be within (-180, -90, 180, 90)."
            )

        self.transform = self.dataset.rio.transform()
        self.height = self.dataset.rio.height
        self.width = self.dataset.rio.width

    def close(self):
        """Close xarray dataset."""
        self._ctx_stack.close()

    def __exit__(self, exc_type, exc_value, traceback):
        """Support using with Context Managers."""
        self.close()

    @property
    def variables(self) -> list[str]:
        """Return dataset variable names"""
        return list(self.dataset.data_vars)  # type: ignore

    def _arrange_dims(self, da: xarray.DataArray) -> xarray.DataArray:
        """Arrange coordinates and time dimensions.

        An rioxarray.exceptions.InvalidDimensionOrder error is raised if the coordinates are not in the correct order time, y, and x.
        See: https://github.com/corteva/rioxarray/discussions/674

        We conform to using x and y as the spatial dimension names..

        """
        if "x" not in da.dims and "y" not in da.dims:
            try:
                latitude_var_name = next(
                    name
                    for name in ["lat", "latitude", "LAT", "LATITUDE", "Lat"]
                    if name in da.dims
                )
                longitude_var_name = next(
                    name
                    for name in ["lon", "longitude", "LON", "LONGITUDE", "Lon"]
                    if name in da.dims
                )
            except StopIteration as e:
                raise ValueError(f"Couldn't find X/Y dimensions in {da.dims}") from e

            da = da.rename({latitude_var_name: "y", longitude_var_name: "x"})

        if "TIME" in da.dims:
            da = da.rename({"TIME": "time"})

        if extra_dims := [d for d in da.dims if d not in ["x", "y"]]:
            da = da.transpose(*extra_dims, "y", "x")
        else:
            da = da.transpose("y", "x")

        # If min/max values are stored in `valid_range` we add them in `valid_min/valid_max`
        vmin, vmax = da.attrs.get("valid_min"), da.attrs.get("valid_max")
        if "valid_range" in da.attrs and not (vmin is not None and vmax is not None):
            valid_range = da.attrs.get("valid_range")
            da.attrs.update({"valid_min": valid_range[0], "valid_max": valid_range[1]})

        return da

    def _get_variable(
        self,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
    ) -> xarray.DataArray:
        """Get DataArray from xarray Dataset."""
        da = self.dataset[variable]

        if sel:
            _idx: dict[str, list] = {}
            for s in sel:
                val: str | slice
                dim, val = s.split("=")

                # cast string to dtype of the dimension
                if da[dim].dtype != "O":
                    val = da[dim].dtype.type(val)

                if dim in _idx:
                    _idx[dim].append(val)
                else:
                    _idx[dim] = [val]

            sel_idx = {k: v[0] if len(v) < 2 else v for k, v in _idx.items()}
            da = da.sel(sel_idx, method=method)

        da = self._arrange_dims(da)

        # Make sure we have a valid CRS
        crs = da.rio.crs or "epsg:4326"
        da = da.rio.write_crs(crs)

        if crs == "epsg:4326" and (da.x > 180).any():
            # Adjust the longitude coordinates to the -180 to 180 range
            da = da.assign_coords(x=(da.x + 180) % 360 - 180)

            # Sort the dataset by the updated longitude coordinates
            da = da.sortby(da.x)

        assert len(da.dims) in [
            2,
            3,
        ], "rio_tiler.io.xarray.DatasetReader can only work with 2D or 3D DataArray"

        return da

    def spatial_info(  # type: ignore
        self,
        *,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
    ):
        """Return xarray.DataArray info."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return {
                "crs": da.crs,
                "bounds": da.bounds,
                "minzoom": da.minzoom,
                "maxzoom": da.maxzoom,
            }

    def get_geographic_bounds(  # type: ignore
        self,
        crs: CRS,
        *,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
    ) -> BBox:
        """Return Geographic Bounds for a Geographic CRS."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.get_geographic_bounds(crs)

    def info(  # type: ignore
        self,
        *,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
    ) -> Info:
        """Return xarray.DataArray info."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.info()

    def statistics(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> dict[str, BandStatistics]:
        """Return statistics from a dataset."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.statistics(*args, **kwargs)

    def tile(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read a Web Map tile from a dataset."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
            tms=self.tms,
        ) as da:
            return da.tile(*args, **kwargs)

    def part(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read part of a dataset."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.part(*args, **kwargs)

    def preview(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Return a preview of a dataset."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.preview(*args, **kwargs)

    def point(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> PointData:
        """Read a pixel value from a dataset."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.point(*args, **kwargs)

    def feature(  # type: ignore
        self,
        *args: Any,
        variable: str,
        sel: list[str] | None = None,
        method: sel_methods | None = None,
        **kwargs: Any,
    ) -> ImageData:
        """Read part of a dataset defined by a geojson feature."""
        with XarrayReader(
            self._get_variable(variable, sel=sel, method=method),
        ) as da:
            return da.feature(*args, **kwargs)



"""rio-tiler types."""

import sys
from collections.abc import Sequence
from typing import Any, Literal, NotRequired

import numpy

if sys.version_info >= (3, 15):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


NumType = float | int

BBox = tuple[float, float, float, float]
NoData = float | int | str
Indexes = Sequence[int] | int

DataMaskType = tuple[numpy.ndarray, numpy.ndarray]

ColorTuple = tuple[int, int, int, int]  # (red, green, blue, alpha)
IntervalTuple = tuple[NumType, NumType]  # (0, 100)

# ColorMap Dict: {1: (0, 0, 0, 255), ...}
GDALColorMapType = dict[int, ColorTuple]

# Discrete Colormap, like GDALColorMapType but accept Float: {0.1: (0, 0, 0, 255), ...}
DiscreteColorMapType = dict[NumType, ColorTuple]

# Intervals ColorMap: [((0, 1), (0, 0, 0, 0)), ...]
IntervalColorMapType = Sequence[tuple[IntervalTuple, ColorTuple]]

ColorMapType = GDALColorMapType | DiscreteColorMapType | IntervalColorMapType

# RasterIO() resampling method.
# ref: https://gdal.org/api/raster_c_api.html#_CPPv418GDALRIOResampleAlg
RIOResampling = Literal[
    "nearest",
    "bilinear",
    "cubic",
    "cubic_spline",
    "lanczos",
    "average",
    "mode",
    "gauss",
    "rms",
]

# WarpKernel resampling method.
# ref: https://gdal.org/en/stable/api/gdalwarp_cpp.html#_CPPv415GDALResampleAlg
WarpResampling = Literal[
    "nearest",
    "bilinear",
    "cubic",
    "cubic_spline",
    "lanczos",
    "average",
    "mode",
    "max",
    "min",
    "med",
    "q1",
    "q3",
    "sum",
    "rms",
]


class AssetWithOptions(TypedDict, extra_items=True):  # type: ignore[call-arg]
    """Asset with additional options."""

    name: str
    indexes: NotRequired[Indexes]
    expression: NotRequired[str]
    bands: NotRequired[Sequence[str]]


AssetType = str | AssetWithOptions


class AssetInfo(TypedDict):
    """Asset Reader Options."""

    url: Any
    name: str
    media_type: str | None
    reader_options: dict
    method_options: dict
    env: NotRequired[dict]
    metadata: NotRequired[dict]
    dataset_statistics: NotRequired[Sequence[tuple[float, float]]]
